"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest

import salt.modules.extfs as extfs
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {extfs: {}}


# 'mkfs' function tests: 1


def test_mkfs():
    """
    Tests if a file system created on the specified device
    """
    mock_ret = {
        "pid": 14247,
        "retcode": 0,
        "stdout": "",
        "stderr": "",
    }
    mock = MagicMock(return_value=mock_ret)
    with patch.dict(extfs.__salt__, {"cmd.run_all": mock}):
        assert extfs.mkfs("/dev/sda1", "ext4") == []
        assert extfs.mkfs("/dev/sda1", "ext4", full_return=True) == {
            "pid": 14247,
            "retcode": 0,
            "stdout": "",
            "stderr": "",
            "comment": [],
        }


# 'tune' function tests: 1


def test_tune():
    """
    Tests if specified group was added
    """
    mock_ret = {
        "pid": 14247,
        "retcode": 1,
        "stdout": "tune2fs 1.44.5 (15-Dec-2018)",
        "stderr": "tune2fs: No such file or directory while trying to open /dev/donkey\nCouldn't find valid filesystem superblock.",
    }
    mock = MagicMock(return_value=mock_ret)
    with patch.dict(extfs.__salt__, {"cmd.run_all": mock}):
        assert extfs.tune("/dev/sda1") == [
            "tune2fs 1.44.5 (15-Dec-2018)",
            "tune2fs: No such file or directory while trying to open /dev/donkey",
            "Couldn't find valid filesystem superblock.",
        ]
        assert extfs.tune("/dev/sda1", full_return=True) == {
            "pid": 14247,
            "retcode": 1,
            "stdout": "tune2fs 1.44.5 (15-Dec-2018)",
            "stderr": "tune2fs: No such file or directory while trying to open /dev/donkey\nCouldn't find valid filesystem superblock.",
            "comment": [
                "tune2fs 1.44.5 (15-Dec-2018)",
                "tune2fs: No such file or directory while trying to open /dev/donkey",
                "Couldn't find valid filesystem superblock.",
            ],
        }


# 'dump' function tests: 1


def test_dump():
    """
    Tests if specified group was added
    """
    mock = MagicMock()
    with patch.dict(extfs.__salt__, {"cmd.run": mock}):
        assert {"attributes": {}, "blocks": {}} == extfs.dump("/dev/sda1")


# 'attributes' function tests: 1


def test_attributes():
    """
    Tests if specified group was added
    """
    with patch(
        "salt.modules.extfs.dump",
        MagicMock(return_value={"attributes": {}, "blocks": {}}),
    ):
        assert {} == extfs.attributes("/dev/sda1")


# 'blocks' function tests: 1


def test_blocks():
    """
    Tests if specified group was added
    """
    with patch(
        "salt.modules.extfs.dump",
        MagicMock(return_value={"attributes": {}, "blocks": {}}),
    ):
        assert {} == extfs.blocks("/dev/sda1")



import logging

import pytest

import salt.modules.file as filemod
from salt.exceptions import SaltInvocationError
from tests.support.mock import MagicMock, patch

log = logging.getLogger(__name__)


@pytest.fixture
def configure_loader_modules():
    return {
        filemod: {
            "__salt__": {},
            "__opts__": {
                "test": False,
                "file_roots": {"base": "tmp"},
                "pillar_roots": {"base": "tmp"},
                "cachedir": "tmp",
                "grains": {},
            },
            "__grains__": {},
            "__utils__": {},
        }
    }


def test_file_rmdir_not_absolute_path_exception():
    with pytest.raises(SaltInvocationError):
        filemod.rmdir("not_absolute")


def test_file_rmdir_not_found_exception():
    with pytest.raises(SaltInvocationError):
        filemod.rmdir("/tmp/not_there")


def test_file_rmdir_success_return():
    with patch("os.rmdir", MagicMock(return_value=True)), patch(
        "os.path.isdir", MagicMock(return_value=True)
    ):
        assert filemod.rmdir("/tmp/salt_test_return") is True


def test_file_rmdir_failure_return():
    with patch(
        "os.rmdir", MagicMock(side_effect=OSError(39, "Directory not empty"))
    ), patch("os.path.isdir", MagicMock(return_value=True)):
        assert filemod.rmdir("/tmp/salt_test_return") is False



"""
    :codeauthor: Rajvi Dhimar <rajvidhimar95@gmail.com>
"""

import os

import pytest
import yaml

import salt.modules.junos as junos
from tests.support.mock import ANY, MagicMock, PropertyMock, call, mock_open, patch

try:
    from lxml import etree
except ImportError:
    import xml.etree.ElementTree as etree

try:
    import jnpr.junos.op as tables_dir
    import jxmlease  # pylint: disable=unused-import
    from jnpr.junos.device import Device
    from jnpr.junos.exception import ConnectClosedError, LockError, UnlockError
    from jnpr.junos.utils.config import Config
    from jnpr.junos.utils.sw import SW

    HAS_JUNOS = True
except ImportError:
    HAS_JUNOS = False

pytestmark = [
    pytest.mark.skip_on_windows(reason="Not supported on Windows"),
    pytest.mark.skipif(
        not HAS_JUNOS, reason="The junos-eznc and jxmlease modules are required"
    ),
]


@pytest.fixture
def get_facts():
    facts = {
        "2RE": True,
        "HOME": "/var/home/regress",
        "RE0": {
            "last_reboot_reason": "0x200:normal shutdown",
            "mastership_state": "master",
            "model": "RE-VMX",
            "status": "OK",
            "up_time": "11 days, 23 hours, 16 minutes, 54 seconds",
        },
        "RE1": {
            "last_reboot_reason": "0x200:normal shutdown",
            "mastership_state": "backup",
            "model": "RE-VMX",
            "status": "OK",
            "up_time": "11 days, 23 hours, 16 minutes, 41 seconds",
        },
        "RE_hw_mi": False,
        "current_re": ["re0", "master", "node", "fwdd", "member", "pfem"],
        "domain": "englab.juniper.net",
        "fqdn": "R1_re0.englab.juniper.net",
        "hostname": "R1_re0",
        "hostname_info": {"re0": "R1_re0", "re1": "R1_re01"},
        "ifd_style": "CLASSIC",
        "junos_info": {
            "re0": {
                "object": {
                    "build": None,
                    "major": (16, 1),
                    "minor": "20160413_0837_aamish",
                    "type": "I",
                },
                "text": "16.1I20160413_0837_aamish",
            },
            "re1": {
                "object": {
                    "build": None,
                    "major": (16, 1),
                    "minor": "20160413_0837_aamish",
                    "type": "I",
                },
                "text": "16.1I20160413_0837_aamish",
            },
        },
        "master": "RE0",
        "model": "MX240",
        "model_info": {"re0": "MX240", "re1": "MX240"},
        "personality": "MX",
        "re_info": {
            "default": {
                "0": {
                    "last_reboot_reason": "0x200:normal shutdown",
                    "mastership_state": "master",
                    "model": "RE-VMX",
                    "status": "OK",
                },
                "1": {
                    "last_reboot_reason": "0x200:normal shutdown",
                    "mastership_state": "backup",
                    "model": "RE-VMX",
                    "status": "OK",
                },
                "default": {
                    "last_reboot_reason": "0x200:normal shutdown",
                    "mastership_state": "master",
                    "model": "RE-VMX",
                    "status": "OK",
                },
            }
        },
        "re_master": {"default": "0"},
        "serialnumber": "VMX4eaf",
        "srx_cluster": None,
        "switch_style": "BRIDGE_DOMAIN",
        "vc_capable": False,
        "vc_fabric": None,
        "vc_master": None,
        "vc_mode": None,
        "version": "16.1I20160413_0837_aamish",
        "version_RE0": "16.1I20160413_0837_aamish",
        "version_RE1": "16.1I20160413_0837_aamish",
        "version_info": {
            "build": None,
            "major": (16, 1),
            "minor": "20160413_0837_aamish",
            "type": "I",
        },
        "virtual": True,
    }
    return facts


@pytest.fixture
def make_connect():
    with patch("ncclient.manager.connect") as mock_connect:
        dev = Device(
            host="1.1.1.1",
            user="test",
            password="test123",
            fact_style="old",
            gather_facts=False,
        )
        dev.open()
        dev.timeout = 30
        dev.bind(cu=Config)
        dev.bind(sw=SW)
        yield dev


@pytest.fixture
def configure_loader_modules(get_facts, make_connect):
    return {
        junos: {
            "__proxy__": {
                "junos.conn": MagicMock(return_value=make_connect),
                "junos.get_serialized_facts": MagicMock(return_value=get_facts),
                "junos.reboot_active": MagicMock(return_value=True),
                "junos.reboot_clear": MagicMock(return_value=True),
            },
            "__salt__": {
                "cp.get_template": MagicMock(return_value=True),
                "cp.get_file": MagicMock(return_value=True),
                "file.file_exists": MagicMock(return_value=True),
                "slsutil.renderer": MagicMock(
                    return_value="set system host-name dummy"
                ),
                "event.fire_master": MagicMock(return_value=None),
            },
            "_restart_connection": MagicMock(return_value=None),
        },
    }


def raise_exception(*args, **kwargs):
    raise Exception("Test exception")


def test__timeout_decorator():
    with patch("jnpr.junos.Device.timeout", new_callable=PropertyMock) as mock_timeout:
        mock_timeout.return_value = 30

        def function(x):
            return x

        decorator = junos._timeout_decorator(function)
        decorator("Test Mock", dev_timeout=10)
        calls = [call(), call(10), call(30)]
        mock_timeout.assert_has_calls(calls)


def test__timeout_cleankwargs_decorator():
    with patch("jnpr.junos.Device.timeout", new_callable=PropertyMock) as mock_timeout:
        mock_timeout.return_value = 30

        def function(x):
            return x

        decorator = junos._timeout_decorator_cleankwargs(function)
        decorator("Test Mock", dev_timeout=10, __pub_args="abc")
        calls = [call(), call(10), call(30)]
        mock_timeout.assert_has_calls(calls)


def test_facts_refresh():
    with patch("salt.modules.saltutil.sync_grains") as mock_sync_grains:
        ret = {
            "out": True,
            "facts": {
                "2RE": True,
                "HOME": "/var/home/regress",
                "RE0": {
                    "last_reboot_reason": "0x200:normal shutdown",
                    "mastership_state": "master",
                    "model": "RE-VMX",
                    "status": "OK",
                    "up_time": "11 days, 23 hours, 16 minutes, 54 seconds",
                },
                "RE1": {
                    "last_reboot_reason": "0x200:normal shutdown",
                    "mastership_state": "backup",
                    "model": "RE-VMX",
                    "status": "OK",
                    "up_time": "11 days, 23 hours, 16 minutes, 41 seconds",
                },
                "RE_hw_mi": False,
                "current_re": ["re0", "master", "node", "fwdd", "member", "pfem"],
                "domain": "englab.juniper.net",
                "fqdn": "R1_re0.englab.juniper.net",
                "hostname": "R1_re0",
                "hostname_info": {"re0": "R1_re0", "re1": "R1_re01"},
                "ifd_style": "CLASSIC",
                "junos_info": {
                    "re0": {
                        "object": {
                            "build": None,
                            "major": (16, 1),
                            "minor": "20160413_0837_aamish",
                            "type": "I",
                        },
                        "text": "16.1I20160413_0837_aamish",
                    },
                    "re1": {
                        "object": {
                            "build": None,
                            "major": (16, 1),
                            "minor": "20160413_0837_aamish",
                            "type": "I",
                        },
                        "text": "16.1I20160413_0837_aamish",
                    },
                },
                "master": "RE0",
                "model": "MX240",
                "model_info": {"re0": "MX240", "re1": "MX240"},
                "personality": "MX",
                "re_info": {
                    "default": {
                        "0": {
                            "last_reboot_reason": "0x200:normal shutdown",
                            "mastership_state": "master",
                            "model": "RE-VMX",
                            "status": "OK",
                        },
                        "1": {
                            "last_reboot_reason": "0x200:normal shutdown",
                            "mastership_state": "backup",
                            "model": "RE-VMX",
                            "status": "OK",
                        },
                        "default": {
                            "last_reboot_reason": "0x200:normal shutdown",
                            "mastership_state": "master",
                            "model": "RE-VMX",
                            "status": "OK",
                        },
                    }
                },
                "re_master": {"default": "0"},
                "serialnumber": "VMX4eaf",
                "srx_cluster": None,
                "switch_style": "BRIDGE_DOMAIN",
                "vc_capable": False,
                "vc_fabric": None,
                "vc_master": None,
                "vc_mode": None,
                "version": "16.1I20160413_0837_aamish",
                "version_RE0": "16.1I20160413_0837_aamish",
                "version_RE1": "16.1I20160413_0837_aamish",
                "version_info": {
                    "build": None,
                    "major": (16, 1),
                    "minor": "20160413_0837_aamish",
                    "type": "I",
                },
                "virtual": True,
            },
        }
        assert junos.facts_refresh() == ret


def test_facts_refresh_exception():
    with patch("jnpr.junos.device.Device.facts_refresh") as mock_facts_refresh:
        mock_facts_refresh.side_effect = raise_exception
        ret = {
            "message": 'Execution failed due to "Test exception"',
            "out": False,
        }
        assert junos.facts_refresh() == ret


def test_facts():
    ret = {
        "out": True,
        "facts": {
            "2RE": True,
            "HOME": "/var/home/regress",
            "RE0": {
                "last_reboot_reason": "0x200:normal shutdown",
                "mastership_state": "master",
                "model": "RE-VMX",
                "status": "OK",
                "up_time": "11 days, 23 hours, 16 minutes, 54 seconds",
            },
            "RE1": {
                "last_reboot_reason": "0x200:normal shutdown",
                "mastership_state": "backup",
                "model": "RE-VMX",
                "status": "OK",
                "up_time": "11 days, 23 hours, 16 minutes, 41 seconds",
            },
            "RE_hw_mi": False,
            "current_re": ["re0", "master", "node", "fwdd", "member", "pfem"],
            "domain": "englab.juniper.net",
            "fqdn": "R1_re0.englab.juniper.net",
            "hostname": "R1_re0",
            "hostname_info": {"re0": "R1_re0", "re1": "R1_re01"},
            "ifd_style": "CLASSIC",
            "junos_info": {
                "re0": {
                    "object": {
                        "build": None,
                        "major": (16, 1),
                        "minor": "20160413_0837_aamish",
                        "type": "I",
                    },
                    "text": "16.1I20160413_0837_aamish",
                },
                "re1": {
                    "object": {
                        "build": None,
                        "major": (16, 1),
                        "minor": "20160413_0837_aamish",
                        "type": "I",
                    },
                    "text": "16.1I20160413_0837_aamish",
                },
            },
            "master": "RE0",
            "model": "MX240",
            "model_info": {"re0": "MX240", "re1": "MX240"},
            "personality": "MX",
            "re_info": {
                "default": {
                    "0": {
                        "last_reboot_reason": "0x200:normal shutdown",
                        "mastership_state": "master",
                        "model": "RE-VMX",
                        "status": "OK",
                    },
                    "1": {
                        "last_reboot_reason": "0x200:normal shutdown",
                        "mastership_state": "backup",
                        "model": "RE-VMX",
                        "status": "OK",
                    },
                    "default": {
                        "last_reboot_reason": "0x200:normal shutdown",
                        "mastership_state": "master",
                        "model": "RE-VMX",
                        "status": "OK",
                    },
                }
            },
            "re_master": {"default": "0"},
            "serialnumber": "VMX4eaf",
            "srx_cluster": None,
            "switch_style": "BRIDGE_DOMAIN",
            "vc_capable": False,
            "vc_fabric": None,
            "vc_master": None,
            "vc_mode": None,
            "version": "16.1I20160413_0837_aamish",
            "version_RE0": "16.1I20160413_0837_aamish",
            "version_RE1": "16.1I20160413_0837_aamish",
            "version_info": {
                "build": None,
                "major": (16, 1),
                "minor": "20160413_0837_aamish",
                "type": "I",
            },
            "virtual": True,
        },
    }
    assert junos.facts() == ret


def test_facts_exception():
    with patch.dict(junos.__proxy__, {"junos.get_serialized_facts": raise_exception}):
        ret = {
            "message": 'Could not display facts due to "Test exception"',
            "out": False,
        }
        assert junos.facts() == ret


def test_set_hostname_without_args():
    ret = {
        "message": "Please provide the hostname.",
        "out": False,
    }
    assert junos.set_hostname() == ret


def test_set_hostname_load_called_with_valid_name():
    with patch("jnpr.junos.utils.config.Config.load") as mock_load:
        junos.set_hostname("test-name")
        mock_load.assert_called_with("set system host-name test-name", format="set")


def test_set_hostname_raise_exception_for_load():
    with patch("jnpr.junos.utils.config.Config.load") as mock_load:
        mock_load.side_effect = raise_exception
        ret = {
            "message": 'Could not load configuration due to error "Test exception"',
            "out": False,
        }
        assert junos.set_hostname("Test-name") == ret


def test_set_hostname_raise_exception_for_commit_check():
    with patch("jnpr.junos.utils.config.Config.commit_check") as mock_commit_check:
        mock_commit_check.side_effect = raise_exception
        ret = {
            "message": 'Could not commit check due to error "Test exception"',
            "out": False,
        }
        assert junos.set_hostname("test-name") == ret


def test_set_hostname_one_arg_parsed_correctly():
    with patch("jnpr.junos.utils.config.Config.load") as mock_load, patch(
        "jnpr.junos.utils.config.Config.commit_check"
    ) as mock_commit_check, patch(
        "jnpr.junos.utils.config.Config.commit"
    ) as mock_commit:
        mock_commit_check.return_value = True
        args = {
            "comment": "Committed via salt",
            "__pub_user": "root",
            "__pub_arg": ["test-name", {"comment": "Committed via salt"}],
            "__pub_fun": "junos.set_hostname",
            "__pub_jid": "20170220210915624885",
            "__pub_tgt": "mac_min",
            "__pub_tgt_type": "glob",
            "__pub_ret": "",
        }

        junos.set_hostname("test-name", **args)
        mock_commit.assert_called_with(comment="Committed via salt")


def test_set_hostname_more_than_one_args_parsed_correctly():
    with patch("jnpr.junos.utils.config.Config.load") as mock_load, patch(
        "jnpr.junos.utils.config.Config.commit_check"
    ) as mock_commit_check, patch(
        "jnpr.junos.utils.config.Config.commit"
    ) as mock_commit:
        mock_commit_check.return_value = True
        args = {
            "comment": "Committed via salt",
            "__pub_user": "root",
            "__pub_arg": [
                "test-name",
                {"comment": "Committed via salt", "confirm": 5},
            ],
            "__pub_fun": "junos.set_hostname",
            "__pub_jid": "20170220210915624885",
            "__pub_tgt": "mac_min",
            "__pub_tgt_type": "glob",
            "__pub_ret": "",
        }

        junos.set_hostname("test-name", **args)
        mock_commit.assert_called_with(comment="Committed via salt", confirm=5)


def test_set_hostname_successful_return_message():
    with patch("jnpr.junos.utils.config.Config.load") as mock_load, patch(
        "jnpr.junos.utils.config.Config.commit_check"
    ) as mock_commit_check, patch(
        "jnpr.junos.utils.config.Config.commit"
    ) as mock_commit:
        mock_commit_check.return_value = True
        args = {
            "comment": "Committed via salt",
            "__pub_user": "root",
            "__pub_arg": ["test-name", {"comment": "Committed via salt"}],
            "__pub_fun": "junos.set_hostname",
            "__pub_jid": "20170220210915624885",
            "__pub_tgt": "mac_min",
            "__pub_tgt_type": "glob",
            "__pub_ret": "",
        }
        ret = {
            "message": "Successfully changed hostname.",
            "out": True,
        }
        assert junos.set_hostname("test-name", **args) == ret


def test_set_hostname_raise_exception_for_commit():
    with patch("jnpr.junos.utils.config.Config.commit") as mock_commit:
        mock_commit.side_effect = raise_exception
        ret = {
            "message": 'Successfully loaded host-name but commit failed with "Test exception"',
            "out": False,
        }
        assert junos.set_hostname("test-name") == ret


def test_set_hostname_fail_commit_check():
    with patch(
        "jnpr.junos.utils.config.Config.commit_check"
    ) as mock_commit_check, patch("salt.modules.junos.rollback") as mock_rollback:
        mock_commit_check.return_value = False
        ret = {
            "message": "Successfully loaded host-name but pre-commit check failed.",
            "out": False,
        }
        assert junos.set_hostname("test") == ret


def test_commit_without_args():
    with patch(
        "jnpr.junos.utils.config.Config.commit_check"
    ) as mock_commit_check, patch(
        "jnpr.junos.utils.config.Config.commit"
    ) as mock_commit:
        mock_commit.return_value = True
        mock_commit_check.return_value = True
        ret = {
            "message": "Commit Successful.",
            "out": True,
        }
        assert junos.commit() == ret


def test_commit_raise_commit_check_exception():
    with patch("jnpr.junos.utils.config.Config.commit_check") as mock_commit_check:
        mock_commit_check.side_effect = raise_exception
        ret = {
            "message": 'Could not perform commit check due to "Test exception"',
            "out": False,
        }
        assert junos.commit() == ret


def test_commit_raise_commit_exception():
    with patch(
        "jnpr.junos.utils.config.Config.commit_check"
    ) as mock_commit_check, patch(
        "jnpr.junos.utils.config.Config.commit"
    ) as mock_commit:
        mock_commit_check.return_value = True
        mock_commit.side_effect = raise_exception
        ret = {
            "message": 'Commit check succeeded but actual commit failed with "Test exception"',
            "out": False,
        }
        assert junos.commit() == ret


def test_commit_with_single_argument():
    with patch(
        "jnpr.junos.utils.config.Config.commit_check"
    ) as mock_commit_check, patch(
        "jnpr.junos.utils.config.Config.commit"
    ) as mock_commit:
        mock_commit_check.return_value = True
        args = {
            "__pub_user": "root",
            "__pub_arg": [{"sync": True}],
            "sync": True,
            "__pub_fun": "junos.commit",
            "__pub_jid": "20170221182531323467",
            "__pub_tgt": "mac_min",
            "__pub_tgt_type": "glob",
            "__pub_ret": "",
        }
        junos.commit(**args)
        mock_commit.assert_called_with(detail=False, sync=True)


def test_commit_with_multiple_arguments():
    with patch(
        "jnpr.junos.utils.config.Config.commit_check"
    ) as mock_commit_check, patch(
        "jnpr.junos.utils.config.Config.commit"
    ) as mock_commit:
        mock_commit_check.return_value = True
        args = {
            "comment": "comitted via salt",
            "__pub_user": "root",
            "__pub_arg": [
                {"comment": "comitted via salt", "confirm": 3, "detail": True}
            ],
            "confirm": 3,
            "detail": True,
            "__pub_fun": "junos.commit",
            "__pub_jid": "20170221182856987820",
            "__pub_tgt": "mac_min",
            "__pub_tgt_type": "glob",
            "__pub_ret": "",
        }
        junos.commit(**args)
        mock_commit.assert_called_with(
            comment="comitted via salt", detail=True, confirm=3
        )


def test_commit_pyez_commit_returning_false():
    with patch(
        "jnpr.junos.utils.config.Config.commit_check"
    ) as mock_commit_check, patch(
        "jnpr.junos.utils.config.Config.commit"
    ) as mock_commit:
        mock_commit.return_value = False
        mock_commit_check.return_value = True
        ret = {
            "message": "Commit failed.",
            "out": False,
        }
        assert junos.commit() == ret


def test_commit_pyez_commit_check_returns_false():
    with patch("jnpr.junos.utils.config.Config.commit_check") as mock_commit_check:
        mock_commit_check.return_value = False
        ret = {
            "message": "Pre-commit check failed.",
            "out": False,
        }
        assert junos.commit() == ret


def test_rollback_exception():
    with patch("jnpr.junos.utils.config.Config.rollback") as mock_rollback:
        mock_rollback.side_effect = raise_exception
        ret = {
            "message": 'Rollback failed due to "Test exception"',
            "out": False,
        }
        assert junos.rollback() == ret


def test_rollback_without_args_success():
    with patch(
        "jnpr.junos.utils.config.Config.commit_check"
    ) as mock_commit_check, patch(
        "jnpr.junos.utils.config.Config.commit"
    ) as mock_commit, patch(
        "jnpr.junos.utils.config.Config.rollback"
    ) as mock_rollback:
        mock_commit_check.return_value = True
        mock_rollback.return_value = True
        ret = {
            "message": "Rollback successful",
            "out": True,
        }
        assert junos.rollback() == ret


def test_rollback_without_args_fail():
    with patch("jnpr.junos.utils.config.Config.rollback") as mock_rollback:
        mock_rollback.return_value = False
        ret = {
            "message": "Rollback failed",
            "out": False,
        }
        assert junos.rollback() == ret


def test_rollback_with_id():
    with patch(
        "jnpr.junos.utils.config.Config.commit_check"
    ) as mock_commit_check, patch(
        "jnpr.junos.utils.config.Config.commit"
    ) as mock_commit, patch(
        "jnpr.junos.utils.config.Config.rollback"
    ) as mock_rollback:
        mock_commit_check.return_value = True
        junos.rollback(id=5)
        mock_rollback.assert_called_with(5)


def test_rollback_with_id_and_single_arg():
    with patch(
        "jnpr.junos.utils.config.Config.commit_check"
    ) as mock_commit_check, patch(
        "jnpr.junos.utils.config.Config.commit"
    ) as mock_commit, patch(
        "jnpr.junos.utils.config.Config.rollback"
    ) as mock_rollback:
        mock_commit_check.return_value = True
        args = {
            "__pub_user": "root",
            "__pub_arg": [2, {"confirm": 2}],
            "confirm": 2,
            "__pub_fun": "junos.rollback",
            "__pub_jid": "20170221184518526067",
            "__pub_tgt": "mac_min",
            "__pub_tgt_type": "glob",
            "__pub_ret": "",
        }
        junos.rollback(id=2, **args)
        mock_rollback.assert_called_with(2)
        mock_commit.assert_called_with(confirm=2)


def test_rollback_with_id_and_multiple_args():
    with patch(
        "jnpr.junos.utils.config.Config.commit_check"
    ) as mock_commit_check, patch(
        "jnpr.junos.utils.config.Config.commit"
    ) as mock_commit, patch(
        "jnpr.junos.utils.config.Config.rollback"
    ) as mock_rollback:
        mock_commit_check.return_value = True
        args = {
            "comment": "Comitted via salt",
            "__pub_user": "root",
            "__pub_arg": [
                2,
                {"comment": "Comitted via salt", "dev_timeout": 40, "confirm": 1},
            ],
            "confirm": 1,
            "__pub_fun": "junos.rollback",
            "__pub_jid": "20170221192708251721",
            "__pub_tgt": "mac_min",
            "__pub_tgt_type": "glob",
            "__pub_ret": "",
        }
        junos.rollback(id=2, **args)
        mock_rollback.assert_called_with(2)
        mock_commit.assert_called_with(
            comment="Comitted via salt", confirm=1, dev_timeout=40
        )


def test_rollback_with_only_single_arg():
    with patch(
        "jnpr.junos.utils.config.Config.commit_check"
    ) as mock_commit_check, patch(
        "jnpr.junos.utils.config.Config.commit"
    ) as mock_commit, patch(
        "jnpr.junos.utils.config.Config.rollback"
    ) as mock_rollback:
        mock_commit_check.return_value = True
        args = {
            "__pub_user": "root",
            "__pub_arg": [{"sync": True}],
            "sync": True,
            "__pub_fun": "junos.rollback",
            "__pub_jid": "20170221193615696475",
            "__pub_tgt": "mac_min",
            "__pub_tgt_type": "glob",
            "__pub_ret": "",
        }
        junos.rollback(**args)
        mock_rollback.assert_called_once_with(0)
        mock_commit.assert_called_once_with(sync=True)


def test_rollback_with_only_multiple_args_no_id():
    with patch(
        "jnpr.junos.utils.config.Config.commit_check"
    ) as mock_commit_check, patch(
        "jnpr.junos.utils.config.Config.commit"
    ) as mock_commit, patch(
        "jnpr.junos.utils.config.Config.rollback"
    ) as mock_rollback:
        mock_commit_check.return_value = True
        args = {
            "comment": "Comitted via salt",
            "__pub_user": "root",
            "__pub_arg": [{"comment": "Comitted via salt", "confirm": 3, "sync": True}],
            "confirm": 3,
            "sync": True,
            "__pub_fun": "junos.rollback",
            "__pub_jid": "20170221193945996362",
            "__pub_tgt": "mac_min",
            "__pub_tgt_type": "glob",
            "__pub_ret": "",
        }
        junos.rollback(**args)
        mock_rollback.assert_called_with(0)
        mock_commit.assert_called_once_with(
            sync=True, confirm=3, comment="Comitted via salt"
        )


def test_rollback_with_diffs_file_option_when_diff_is_None():
    with patch(
        "jnpr.junos.utils.config.Config.commit_check"
    ) as mock_commit_check, patch(
        "jnpr.junos.utils.config.Config.commit"
    ) as mock_commit, patch(
        "jnpr.junos.utils.config.Config.rollback"
    ) as mock_rollback, patch(
        "salt.utils.files.fopen"
    ) as mock_fopen, patch(
        "jnpr.junos.utils.config.Config.diff"
    ) as mock_diff:
        mock_commit_check.return_value = True
        mock_diff.return_value = "diff"
        args = {
            "__pub_user": "root",
            "__pub_arg": [{"diffs_file": "/home/regress/diff", "confirm": 2}],
            "confirm": 2,
            "__pub_fun": "junos.rollback",
            "__pub_jid": "20170221205153884009",
            "__pub_tgt": "mac_min",
            "__pub_tgt_type": "glob",
            "__pub_ret": "",
            "diffs_file": "/home/regress/diff",
        }
        junos.rollback(**args)
        mock_fopen.assert_called_with("/home/regress/diff", "w")


def test_rollback_with_diffs_file_option():
    with patch(
        "jnpr.junos.utils.config.Config.commit_check"
    ) as mock_commit_check, patch(
        "jnpr.junos.utils.config.Config.commit"
    ) as mock_commit, patch(
        "jnpr.junos.utils.config.Config.rollback"
    ) as mock_rollback, patch(
        "salt.utils.files.fopen"
    ) as mock_fopen, patch(
        "jnpr.junos.utils.config.Config.diff"
    ) as mock_diff:
        mock_commit_check.return_value = True
        mock_diff.return_value = None
        args = {
            "__pub_user": "root",
            "__pub_arg": [{"diffs_file": "/home/regress/diff", "confirm": 2}],
            "confirm": 2,
            "__pub_fun": "junos.rollback",
            "__pub_jid": "20170221205153884009",
            "__pub_tgt": "mac_min",
            "__pub_tgt_type": "glob",
            "__pub_ret": "",
            "diffs_file": "/home/regress/diff",
        }
        junos.rollback(**args)
        assert not mock_fopen.called


def test_rollback_commit_check_exception():
    with patch(
        "jnpr.junos.utils.config.Config.commit_check"
    ) as mock_commit_check, patch(
        "jnpr.junos.utils.config.Config.rollback"
    ) as mock_rollback:
        mock_commit_check.side_effect = raise_exception
        ret = {
            "message": 'Could not commit check due to "Test exception"',
            "out": False,
        }
        assert junos.rollback() == ret


def test_rollback_commit_exception():
    with patch(
        "jnpr.junos.utils.config.Config.commit_check"
    ) as mock_commit_check, patch(
        "jnpr.junos.utils.config.Config.commit"
    ) as mock_commit, patch(
        "jnpr.junos.utils.config.Config.rollback"
    ) as mock_rollback:
        mock_commit_check.return_value = True
        mock_commit.side_effect = raise_exception
        ret = {
            "message": 'Rollback successful but commit failed with error "Test exception"',
            "out": False,
        }
        assert junos.rollback() == ret


def test_rollback_commit_check_fails():
    with patch(
        "jnpr.junos.utils.config.Config.commit_check"
    ) as mock_commit_check, patch(
        "jnpr.junos.utils.config.Config.rollback"
    ) as mock_rollback:
        mock_commit_check.return_value = False
        ret = {
            "message": "Rollback successful but pre-commit check failed.",
            "out": False,
        }
        assert junos.rollback() == ret


def test_diff_without_args():
    with patch("jnpr.junos.utils.config.Config.diff") as mock_diff:
        junos.diff()
        mock_diff.assert_called_with(rb_id=0)


def test_diff_with_arg():
    with patch("jnpr.junos.utils.config.Config.diff") as mock_diff:
        junos.diff(id=2)
        mock_diff.assert_called_with(rb_id=2)


def test_diff_exception():
    with patch("jnpr.junos.utils.config.Config.diff") as mock_diff:
        mock_diff.side_effect = raise_exception
        ret = {
            "message": 'Could not get diff with error "Test exception"',
            "out": False,
        }
        assert junos.diff() == ret


def test_ping_without_args():
    ret = {
        "message": "Please specify the destination ip to ping.",
        "out": False,
    }
    assert junos.ping() == ret


def test_ping():
    with patch("jnpr.junos.device.Device.execute") as mock_execute:
        junos.ping("1.1.1.1")
        args = mock_execute.call_args
        rpc = b"<ping><host>1.1.1.1</host><count>5</count></ping>"
        mydgm = etree.tostring(args[0][0])
        assert etree.tostring(args[0][0]) == rpc


def test_ping_ttl():
    with patch("jnpr.junos.device.Device.execute") as mock_execute:
        args = {
            "__pub_user": "sudo_drajvi",
            "__pub_arg": ["1.1.1.1", {"ttl": 3}],
            "__pub_fun": "junos.ping",
            "__pub_jid": "20170306165237683279",
            "__pub_tgt": "mac_min",
            "ttl": 3,
            "__pub_tgt_type": "glob",
            "__pub_ret": "",
        }
        junos.ping("1.1.1.1", **args)
        exec_args = mock_execute.call_args
        rpc = b"<ping><host>1.1.1.1</host><ttl>3</ttl><count>5</count></ping>"
        assert etree.tostring(exec_args[0][0]) == rpc


def test_ping_exception():
    with patch("jnpr.junos.device.Device.execute") as mock_execute:
        mock_execute.side_effect = raise_exception
        ret = {
            "message": 'Execution failed due to "Test exception"',
            "out": False,
        }
        assert junos.ping("1.1.1.1") == ret


def test_cli_without_args():
    ret = {
        "message": "Please provide the CLI command to be executed.",
        "out": False,
    }
    assert junos.cli() == ret


def test_cli_with_format_as_empty_string():
    with patch("jnpr.junos.device.Device.cli") as mock_cli:
        junos.cli("show version", format="")
        mock_cli.assert_called_with("show version", "text", warning=False)


def test_cli():
    with patch("jnpr.junos.device.Device.cli") as mock_cli:
        mock_cli.return_vale = "CLI result"
        ret = {
            "message": "CLI result",
            "out": True,
        }
        junos.cli("show version")
        mock_cli.assert_called_with("show version", "text", warning=False)


def test_cli_format_xml():
    with patch("salt.modules.junos.jxmlease.parse") as mock_jxml, patch(
        "salt.modules.junos.etree.tostring"
    ) as mock_to_string, patch("jnpr.junos.device.Device.cli") as mock_cli:
        mock_cli.return_value = "<root><a>test</a></root>"
        mock_jxml.return_value = "<root><a>test</a></root>"
        args = {
            "__pub_user": "root",
            "__pub_arg": [{"format": "xml"}],
            "format": "xml",
            "__pub_fun": "junos.cli",
            "__pub_jid": "20170221182531323467",
            "__pub_tgt": "mac_min",
            "__pub_tgt_type": "glob",
            "__pub_ret": "",
        }
        ret = {
            "message": "<root><a>test</a></root>",
            "out": True,
        }
        assert junos.cli("show version", **args) == ret
        mock_cli.assert_called_with("show version", "xml", warning=False)
        mock_to_string.assert_called_once_with("<root><a>test</a></root>")
        assert mock_jxml.called


def test_cli_exception_in_cli():
    with patch("jnpr.junos.device.Device.cli") as mock_cli:
        mock_cli.side_effect = raise_exception
        ret = {
            "message": 'Execution failed due to "Test exception"',
            "out": False,
        }
        assert junos.cli("show version") == ret


def test_cli_output_save():
    with patch("jnpr.junos.device.Device.cli") as mock_cli, patch(
        "salt.utils.files.fopen"
    ) as mock_fopen:
        mock_cli.return_value = "Test return"
        args = {
            "__pub_user": "root",
            "__pub_arg": [{"format": "text", "dest": "/path/to/file"}],
            "format": "text",
            "dest": "/path/to/file",
            "__pub_fun": "junos.cli",
            "__pub_jid": "20170221182531323467",
            "__pub_tgt": "mac_min",
            "__pub_tgt_type": "glob",
            "__pub_ret": "",
        }
        ret = {
            "message": "Test return",
            "out": True,
        }
        assert junos.cli("show version", **args) == ret
        mock_fopen.assert_called_with("/path/to/file", "w")
        mock_cli.assert_called_with("show version", "text", warning=False)


def test_cli_output_save_ioexception():
    with patch("jnpr.junos.device.Device.cli") as mock_cli, patch(
        "salt.utils.files.fopen"
    ) as mock_fopen:
        mock_cli.return_value = "Test return"
        mock_fopen.side_effect = IOError()
        args = {
            "__pub_user": "root",
            "__pub_arg": [{"format": "text", "dest": "/path/to/file"}],
            "format": "text",
            "dest": "/path/to/file",
            "__pub_fun": "junos.cli",
            "__pub_jid": "20170221182531323467",
            "__pub_tgt": "mac_min",
            "__pub_tgt_type": "glob",
            "__pub_ret": "",
        }
        ret = {
            "message": 'Unable to open "/path/to/file" to write',
            "out": False,
        }
        assert junos.cli("show version", **args) == ret


def test_shutdown_without_args():
    ret = {
        "message": "Provide either one of the arguments: shutdown or reboot.",
        "out": False,
    }
    assert junos.shutdown() == ret


def test_shutdown_with_reboot_args():
    with patch("salt.modules.junos.SW.reboot") as mock_reboot:
        ret = {
            "message": "Successfully powered off/rebooted.",
            "out": True,
        }
        args = {
            "__pub_user": "root",
            "__pub_arg": [{"reboot": True}],
            "reboot": True,
            "__pub_fun": "junos.shutdown",
            "__pub_jid": "20170222213858582619",
            "__pub_tgt": "mac_min",
            "__pub_tgt_type": "glob",
            "__pub_ret": "",
        }
        assert junos.shutdown(**args) == ret
        assert mock_reboot.called


def test_shutdown_with_poweroff_args():
    with patch("salt.modules.junos.SW.poweroff") as mock_poweroff:
        ret = {
            "message": "Successfully powered off/rebooted.",
            "out": True,
        }
        args = {
            "__pub_user": "root",
            "__pub_arg": [{"shutdown": True}],
            "reboot": True,
            "__pub_fun": "junos.shutdown",
            "__pub_jid": "20170222213858582619",
            "__pub_tgt": "mac_min",
            "__pub_tgt_type": "glob",
            "__pub_ret": "",
        }
        assert junos.shutdown(**args) == ret
        assert mock_poweroff.called


def test_shutdown_with_shutdown_as_false():
    ret = {
        "message": "Nothing to be done.",
        "out": False,
    }
    args = {
        "__pub_user": "root",
        "__pub_arg": [{"shutdown": False}],
        "reboot": True,
        "__pub_fun": "junos.shutdown",
        "__pub_jid": "20170222213858582619",
        "__pub_tgt": "mac_min",
        "__pub_tgt_type": "glob",
        "__pub_ret": "",
    }
    assert junos.shutdown(**args) == ret


def test_shutdown_with_in_min_arg():
    with patch("salt.modules.junos.SW.poweroff") as mock_poweroff:
        args = {
            "__pub_user": "root",
            "in_min": 10,
            "__pub_arg": [{"in_min": 10, "shutdown": True}],
            "reboot": True,
            "__pub_fun": "junos.shutdown",
            "__pub_jid": "20170222231445709212",
            "__pub_tgt": "mac_min",
            "__pub_tgt_type": "glob",
            "__pub_ret": "",
        }
        junos.shutdown(**args)
        mock_poweroff.assert_called_with(in_min=10)


def test_shutdown_with_at_arg():
    with patch("salt.modules.junos.SW.reboot") as mock_reboot:
        args = {
            "__pub_user": "root",
            "__pub_arg": [{"at": "12:00 pm", "reboot": True}],
            "reboot": True,
            "__pub_fun": "junos.shutdown",
            "__pub_jid": "201702276857",
            "at": "12:00 pm",
            "__pub_tgt": "mac_min",
            "__pub_tgt_type": "glob",
            "__pub_ret": "",
        }
        junos.shutdown(**args)
        mock_reboot.assert_called_with(at="12:00 pm")


def test_shutdown_fail_with_exception():
    with patch("salt.modules.junos.SW.poweroff") as mock_poweroff:
        mock_poweroff.side_effect = raise_exception
        args = {
            "__pub_user": "root",
            "__pub_arg": [{"shutdown": True}],
            "shutdown": True,
            "__pub_fun": "junos.shutdown",
            "__pub_jid": "20170222213858582619",
            "__pub_tgt": "mac_min",
            "__pub_tgt_type": "glob",
            "__pub_ret": "",
        }
        ret = {
            "message": 'Could not poweroff/reboot because "Test exception"',
            "out": False,
        }
        assert junos.shutdown(**args) == ret


def test_install_config_without_args():
    ret = {
        "message": "Please provide the salt path where the configuration is present",
        "out": False,
    }
    assert junos.install_config() == ret


def test_install_config_cp_fails():
    with patch.dict(
        junos.__salt__, {"file.file_exists": MagicMock(return_value=False)}
    ):
        ret = {
            "message": "Invalid file path.",
            "out": False,
        }
        assert junos.install_config("path") == ret


def test_install_config_file_cp_fails():
    with patch.dict(
        junos.__salt__, {"file.file_exists": MagicMock(return_value=False)}
    ):
        ret = {
            "message": "Invalid file path.",
            "out": False,
        }
        assert junos.install_config("path") == ret


def test_install_config():
    with patch.dict(
        junos.__salt__,
        {
            "cp.is_cached": MagicMock(return_value="test/path/config"),
            "cp.hash_file": MagicMock(
                return_value={"hash_type": "sha256", "hsum": "a386e49c17"}
            ),
            "file.get_hash": MagicMock(return_value="a386e49c17"),
        },
    ):
        with patch("jnpr.junos.utils.config.Config.commit") as mock_commit, patch(
            "jnpr.junos.utils.config.Config.commit_check"
        ) as mock_commit_check, patch(
            "jnpr.junos.utils.config.Config.diff"
        ) as mock_diff, patch(
            "jnpr.junos.utils.config.Config.load"
        ) as mock_load, patch(
            "salt.utils.files.safe_rm"
        ) as mock_safe_rm, patch(
            "salt.utils.files.mkstemp"
        ) as mock_mkstemp, patch(
            "salt.utils.files.fopen", mock_open(), create=True
        ) as m_open, patch(
            "os.path.isfile"
        ) as mock_isfile, patch(
            "os.path.getsize"
        ) as mock_getsize:
            mock_isfile.return_value = True
            mock_getsize.return_value = 10
            mock_mkstemp.return_value = "test/path/config"
            mock_diff.return_value = "diff"
            mock_commit_check.return_value = True

            ret = {
                "message": "Successfully loaded and committed!",
                "out": True,
            }
            assert junos.install_config("salt://actual/path/config.set") == ret
            mock_load.assert_called_with(path="test/path/config", format="set")


def test_install_config_xml_file():
    with patch.dict(
        junos.__salt__,
        {
            "cp.is_cached": MagicMock(return_value="test/path/config"),
            "cp.hash_file": MagicMock(
                return_value={"hash_type": "sha256", "hsum": "a386e49c17"}
            ),
            "file.get_hash": MagicMock(return_value="a386e49c17"),
        },
    ):
        with patch("jnpr.junos.utils.config.Config.commit") as mock_commit, patch(
            "jnpr.junos.utils.config.Config.commit_check"
        ) as mock_commit_check, patch(
            "jnpr.junos.utils.config.Config.diff"
        ) as mock_diff, patch(
            "jnpr.junos.utils.config.Config.load"
        ) as mock_load, patch(
            "salt.utils.files.safe_rm"
        ) as mock_safe_rm, patch(
            "salt.utils.files.mkstemp"
        ) as mock_mkstemp, patch(
            "os.path.isfile"
        ) as mock_isfile, patch(
            "salt.utils.files.fopen", mock_open(), create=True
        ) as m_open, patch(
            "os.path.getsize"
        ) as mock_getsize:
            mock_isfile.return_value = True
            mock_getsize.return_value = 10
            mock_mkstemp.return_value = "test/path/config"
            mock_diff.return_value = "diff"
            mock_commit_check.return_value = True

            ret = {
                "message": "Successfully loaded and committed!",
                "out": True,
            }
            assert junos.install_config("salt://actual/path/config.xml") == ret
            mock_load.assert_called_with(path="test/path/config", format="xml")


def test_install_config_text_file():
    with patch.dict(
        junos.__salt__,
        {
            "cp.is_cached": MagicMock(return_value="test/path/config"),
            "cp.hash_file": MagicMock(
                return_value={"hash_type": "sha256", "hsum": "a386e49c17"}
            ),
            "file.get_hash": MagicMock(return_value="a386e49c17"),
        },
    ):
        with patch("jnpr.junos.utils.config.Config.commit") as mock_commit, patch(
            "jnpr.junos.utils.config.Config.commit_check"
        ) as mock_commit_check, patch(
            "jnpr.junos.utils.config.Config.diff"
        ) as mock_diff, patch(
            "jnpr.junos.utils.config.Config.load"
        ) as mock_load, patch(
            "salt.utils.files.safe_rm"
        ) as mock_safe_rm, patch(
            "salt.utils.files.mkstemp"
        ) as mock_mkstemp, patch(
            "os.path.isfile"
        ) as mock_isfile, patch(
            "salt.utils.files.fopen", mock_open(), create=True
        ) as m_open, patch(
            "os.path.getsize"
        ) as mock_getsize:
            mock_isfile.return_value = True
            mock_getsize.return_value = 10
            mock_mkstemp.return_value = "test/path/config"
            mock_diff.return_value = "diff"
            mock_commit_check.return_value = True

            ret = {
                "message": "Successfully loaded and committed!",
                "out": True,
            }
            assert junos.install_config("salt://actual/path/config") == ret
            mock_load.assert_called_with(path="test/path/config", format="text")


def test_install_config_cache_not_exists():
    with patch.dict(
        junos.__salt__,
        {
            "cp.is_cached": MagicMock(return_value=None),
            "file.rmdir": MagicMock(return_value="True"),
        },
    ):
        with patch("jnpr.junos.utils.config.Config.commit") as mock_commit, patch(
            "jnpr.junos.utils.config.Config.commit_check"
        ) as mock_commit_check, patch(
            "jnpr.junos.utils.config.Config.diff"
        ) as mock_diff, patch(
            "jnpr.junos.utils.config.Config.load"
        ) as mock_load, patch(
            "salt.utils.files.safe_rm"
        ) as mock_safe_rm, patch(
            "salt.utils.files.mkstemp"
        ) as mock_mkstemp, patch(
            "tempfile.mkdtemp"
        ) as mock_mkdtemp, patch(
            "os.path.isfile"
        ) as mock_isfile, patch(
            "salt.utils.files.fopen", mock_open(), create=True
        ) as m_open, patch(
            "os.path.getsize"
        ) as mock_getsize:
            mock_isfile.return_value = True
            mock_getsize.return_value = 10
            mock_mkstemp.return_value = "test/path/config"
            mock_diff.return_value = "diff"
            mock_commit_check.return_value = True
            mock_mkdtemp.return_value = "/tmp/argr5351afd"

            ret = {
                "message": "Successfully loaded and committed!",
                "out": True,
            }
            assert (
                junos.install_config("salt://actual/path/config", template_vars=True)
                == ret
            )
            mock_mkstemp.assert_called_with()


def test_install_config_replace():
    with patch.dict(
        junos.__salt__,
        {
            "cp.is_cached": MagicMock(return_value="test/path/config"),
            "cp.hash_file": MagicMock(
                return_value={"hash_type": "sha256", "hsum": "a386e49c17"}
            ),
            "file.get_hash": MagicMock(return_value="a386e49c17"),
        },
    ):
        with patch("jnpr.junos.utils.config.Config.commit") as mock_commit, patch(
            "jnpr.junos.utils.config.Config.commit_check"
        ) as mock_commit_check, patch(
            "jnpr.junos.utils.config.Config.diff"
        ) as mock_diff, patch(
            "jnpr.junos.utils.config.Config.load"
        ) as mock_load, patch(
            "salt.utils.files.safe_rm"
        ) as mock_safe_rm, patch(
            "salt.utils.files.mkstemp"
        ) as mock_mkstemp, patch(
            "os.path.isfile"
        ) as mock_isfile, patch(
            "salt.utils.files.fopen", mock_open(), create=True
        ) as m_open, patch(
            "os.path.getsize"
        ) as mock_getsize:
            mock_isfile.return_value = True
            mock_getsize.return_value = 10
            mock_mkstemp.return_value = "test/path/config"
            mock_diff.return_value = "diff"
            mock_commit_check.return_value = True

            args = {
                "__pub_user": "root",
                "__pub_arg": [{"replace": True}],
                "replace": True,
                "__pub_fun": "junos.install_config",
                "__pub_jid": "20170222213858582619",
                "__pub_tgt": "mac_min",
                "__pub_tgt_type": "glob",
                "__pub_ret": "",
            }

            ret = {
                "message": "Successfully loaded and committed!",
                "out": True,
            }
            assert junos.install_config("salt://actual/path/config.set", **args) == ret
            mock_load.assert_called_with(
                path="test/path/config", format="set", merge=False
            )


def test_install_config_overwrite():
    with patch.dict(
        junos.__salt__,
        {
            "cp.is_cached": MagicMock(return_value="test/path/config"),
            "cp.hash_file": MagicMock(
                return_value={"hash_type": "sha256", "hsum": "a386e49c17"}
            ),
            "file.get_hash": MagicMock(return_value="a386e49c17"),
        },
    ):
        with patch("jnpr.junos.utils.config.Config.commit") as mock_commit, patch(
            "jnpr.junos.utils.config.Config.commit_check"
        ) as mock_commit_check, patch(
            "jnpr.junos.utils.config.Config.diff"
        ) as mock_diff, patch(
            "jnpr.junos.utils.config.Config.load"
        ) as mock_load, patch(
            "salt.utils.files.safe_rm"
        ) as mock_safe_rm, patch(
            "salt.utils.files.mkstemp"
        ) as mock_mkstemp, patch(
            "os.path.isfile"
        ) as mock_isfile, patch(
            "salt.utils.files.fopen", mock_open(), create=True
        ) as m_open, patch(
            "os.path.getsize"
        ) as mock_getsize:
            mock_isfile.return_value = True
            mock_getsize.return_value = 10
            mock_mkstemp.return_value = "test/path/config"
            mock_diff.return_value = "diff"
            mock_commit_check.return_value = True

            args = {
                "__pub_user": "root",
                "__pub_arg": [{"overwrite": True}],
                "overwrite": True,
                "__pub_fun": "junos.install_config",
                "__pub_jid": "20170222213858582619",
                "__pub_tgt": "mac_min",
                "__pub_tgt_type": "glob",
                "__pub_ret": "",
            }

            ret = {
                "message": "Successfully loaded and committed!",
                "out": True,
            }
            assert junos.install_config("salt://actual/path/config.xml", **args) == ret
            mock_load.assert_called_with(
                path="test/path/config", format="xml", overwrite=True
            )


def test_install_config_overwrite_false():
    with patch.dict(
        junos.__salt__,
        {
            "cp.is_cached": MagicMock(return_value="test/path/config"),
            "cp.hash_file": MagicMock(
                return_value={"hash_type": "sha256", "hsum": "a386e49c17"}
            ),
            "file.get_hash": MagicMock(return_value="a386e49c17"),
        },
    ):
        with patch("jnpr.junos.utils.config.Config.commit") as mock_commit, patch(
            "jnpr.junos.utils.config.Config.commit_check"
        ) as mock_commit_check, patch(
            "jnpr.junos.utils.config.Config.diff"
        ) as mock_diff, patch(
            "jnpr.junos.utils.config.Config.load"
        ) as mock_load, patch(
            "salt.utils.files.safe_rm"
        ) as mock_safe_rm, patch(
            "salt.utils.files.mkstemp"
        ) as mock_mkstemp, patch(
            "os.path.isfile"
        ) as mock_isfile, patch(
            "salt.utils.files.fopen", mock_open(), create=True
        ) as m_open, patch(
            "os.path.getsize"
        ) as mock_getsize:
            mock_isfile.return_value = True
            mock_getsize.return_value = 10
            mock_mkstemp.return_value = "test/path/config"
            mock_diff.return_value = "diff"
            mock_commit_check.return_value = True

            args = {
                "__pub_user": "root",
                "__pub_arg": [{"overwrite": False}],
                "overwrite": False,
                "__pub_fun": "junos.install_config",
                "__pub_jid": "20170222213858582619",
                "__pub_tgt": "mac_min",
                "__pub_tgt_type": "glob",
                "__pub_ret": "",
            }

            ret = {
                "message": "Successfully loaded and committed!",
                "out": True,
            }
            assert junos.install_config("salt://actual/path/config", **args) == ret
            mock_load.assert_called_with(
                path="test/path/config", format="text", merge=True
            )


def test_install_config_load_causes_exception():
    with patch("jnpr.junos.utils.config.Config.diff") as mock_diff, patch(
        "jnpr.junos.utils.config.Config.load"
    ) as mock_load, patch("salt.utils.files.safe_rm") as mock_safe_rm, patch(
        "salt.utils.files.mkstemp"
    ) as mock_mkstemp, patch(
        "os.path.isfile"
    ) as mock_isfile, patch(
        "salt.utils.files.fopen", mock_open(), create=True
    ) as m_open, patch(
        "os.path.getsize"
    ) as mock_getsize:
        mock_isfile.return_value = True
        mock_getsize.return_value = 10
        mock_mkstemp.return_value = "test/path/config"
        mock_load.side_effect = raise_exception
        ret = {
            "message": 'Could not load configuration due to : "Test exception"',
            "out": False,
            "format": "set",
        }
        assert junos.install_config(path="actual/path/config.set") == ret


def test_install_config_no_diff():
    with patch("jnpr.junos.utils.config.Config.diff") as mock_diff, patch(
        "jnpr.junos.utils.config.Config.load"
    ) as mock_load, patch("salt.utils.files.safe_rm") as mock_safe_rm, patch(
        "salt.utils.files.mkstemp"
    ) as mock_mkstemp, patch(
        "os.path.isfile"
    ) as mock_isfile, patch(
        "salt.utils.files.fopen", mock_open(), create=True
    ) as m_open, patch(
        "os.path.getsize"
    ) as mock_getsize:
        mock_isfile.return_value = True
        mock_getsize.return_value = 10
        mock_mkstemp.return_value = "test/path/config"
        mock_diff.return_value = None
        ret = {
            "message": "Configuration already applied!",
            "out": True,
        }
        assert junos.install_config("actual/path/config") == ret


def test_install_config_write_diff():
    with patch("jnpr.junos.utils.config.Config.commit") as mock_commit, patch(
        "jnpr.junos.utils.config.Config.commit_check"
    ) as mock_commit_check, patch(
        "jnpr.junos.utils.config.Config.diff"
    ) as mock_diff, patch(
        "jnpr.junos.utils.config.Config.load"
    ) as mock_load, patch(
        "salt.utils.files.safe_rm"
    ) as mock_safe_rm, patch(
        "salt.utils.files.mkstemp"
    ) as mock_mkstemp, patch(
        "os.path.isfile"
    ) as mock_isfile, patch(
        "salt.utils.files.fopen"
    ) as mock_fopen, patch(
        "os.path.getsize"
    ) as mock_getsize:
        mock_isfile.return_value = True
        mock_getsize.return_value = 10
        mock_mkstemp.return_value = "test/path/config"
        mock_diff.return_value = "diff"
        mock_commit_check.return_value = True

        args = {
            "__pub_user": "root",
            "__pub_arg": [{"diffs_file": "copy/config/here"}],
            "diffs_file": "copy/config/here",
            "__pub_fun": "junos.install_config",
            "__pub_jid": "20170222213858582619",
            "__pub_tgt": "mac_min",
            "__pub_tgt_type": "glob",
            "__pub_ret": "",
        }

        ret = {
            "message": "Successfully loaded and committed!",
            "out": True,
        }
        assert junos.install_config("actual/path/config", **args) == ret
        mock_fopen.assert_called_with("copy/config/here", "w")


def test_install_config_write_diff_exception():
    with patch("jnpr.junos.utils.config.Config.commit") as mock_commit, patch(
        "jnpr.junos.utils.config.Config.commit_check"
    ) as mock_commit_check, patch(
        "jnpr.junos.utils.config.Config.diff"
    ) as mock_diff, patch(
        "jnpr.junos.utils.config.Config.load"
    ) as mock_load, patch(
        "salt.utils.files.safe_rm"
    ) as mock_safe_rm, patch(
        "salt.utils.files.mkstemp"
    ) as mock_mkstemp, patch(
        "os.path.isfile"
    ) as mock_isfile, patch(
        "salt.utils.files.fopen", mock_open(), create=True
    ) as mock_fopen, patch(
        "salt.utils.stringutils.to_str"
    ) as mock_strgutils, patch(
        "os.path.getsize"
    ) as mock_getsize:
        mock_isfile.return_value = True
        mock_getsize.return_value = 10
        mock_mkstemp.return_value = "test/path/config"
        mock_diff.return_value = "diff"
        mock_commit_check.return_value = True
        mock_strgutils.side_effect = raise_exception

        args = {
            "__pub_user": "root",
            "__pub_arg": [{"diffs_file": "copy/config/here"}],
            "diffs_file": "copy/config/here",
            "__pub_fun": "junos.install_config",
            "__pub_jid": "20170222213858582619",
            "__pub_tgt": "mac_min",
            "__pub_tgt_type": "glob",
            "__pub_ret": "",
        }

        ret = {
            "message": "Could not write into diffs_file due to: 'Test exception'",
            "out": False,
        }
        assert junos.install_config("actual/path/config", **args) == ret


def test_install_config_commit_params():
    with patch("jnpr.junos.utils.config.Config.commit") as mock_commit, patch(
        "jnpr.junos.utils.config.Config.commit_check"
    ) as mock_commit_check, patch(
        "jnpr.junos.utils.config.Config.diff"
    ) as mock_diff, patch(
        "jnpr.junos.utils.config.Config.load"
    ) as mock_load, patch(
        "salt.utils.files.safe_rm"
    ) as mock_safe_rm, patch(
        "salt.utils.files.mkstemp"
    ) as mock_mkstemp, patch(
        "os.path.isfile"
    ) as mock_isfile, patch(
        "salt.utils.files.fopen", mock_open(), create=True
    ) as m_open, patch(
        "os.path.getsize"
    ) as mock_getsize:
        mock_isfile.return_value = True
        mock_getsize.return_value = 10
        mock_mkstemp.return_value = "test/path/config"
        mock_diff.return_value = "diff"
        mock_commit_check.return_value = True
        args = {
            "comment": "comitted via salt",
            "__pub_user": "root",
            "__pub_arg": [{"comment": "comitted via salt", "confirm": 3}],
            "confirm": 3,
            "__pub_fun": "junos.commit",
            "__pub_jid": "20170221182856987820",
            "__pub_tgt": "mac_min",
            "__pub_tgt_type": "glob",
            "__pub_ret": "",
        }
        ret = {
            "message": "Successfully loaded and committed!",
            "out": True,
        }
        assert junos.install_config("actual/path/config", **args) == ret
        mock_commit.assert_called_with(comment="comitted via salt", confirm=3)


def test_install_config_commit_check_fails():
    with patch(
        "jnpr.junos.utils.config.Config.commit_check"
    ) as mock_commit_check, patch(
        "jnpr.junos.utils.config.Config.diff"
    ) as mock_diff, patch(
        "jnpr.junos.utils.config.Config.load"
    ) as mock_load, patch(
        "salt.utils.files.safe_rm"
    ) as mock_safe_rm, patch(
        "salt.utils.files.mkstemp"
    ) as mock_mkstemp, patch(
        "os.path.isfile"
    ) as mock_isfile, patch(
        "salt.utils.files.fopen", mock_open(), create=True
    ) as m_open, patch(
        "os.path.getsize"
    ) as mock_getsize:
        mock_isfile.return_value = True
        mock_getsize.return_value = 10
        mock_mkstemp.return_value = "test/path/config"
        mock_diff.return_value = "diff"
        mock_commit_check.return_value = False

        ret = {
            "message": "Loaded configuration but commit check failed, hence rolling back configuration.",
            "out": False,
        }
        assert junos.install_config("actual/path/config.xml") == ret


def test_install_config_commit_exception():
    with patch("jnpr.junos.utils.config.Config.commit") as mock_commit, patch(
        "jnpr.junos.utils.config.Config.commit_check"
    ) as mock_commit_check, patch(
        "jnpr.junos.utils.config.Config.diff"
    ) as mock_diff, patch(
        "jnpr.junos.utils.config.Config.load"
    ) as mock_load, patch(
        "salt.utils.files.safe_rm"
    ) as mock_safe_rm, patch(
        "salt.utils.files.mkstemp"
    ) as mock_mkstemp, patch(
        "os.path.isfile"
    ) as mock_isfile, patch(
        "salt.utils.files.fopen", mock_open(), create=True
    ) as m_open, patch(
        "os.path.getsize"
    ) as mock_getsize:
        mock_isfile.return_value = True
        mock_getsize.return_value = 10
        mock_mkstemp.return_value = "test/path/config"
        mock_diff.return_value = "diff"
        mock_commit_check.return_value = True
        mock_commit.side_effect = raise_exception
        ret = {
            "message": 'Commit check successful but commit failed with "Test exception"',
            "out": False,
        }
        assert junos.install_config("actual/path/config") == ret


def test_install_config_test_mode():
    with patch("jnpr.junos.utils.config.Config.commit") as mock_commit, patch(
        "jnpr.junos.utils.config.Config.commit_check"
    ) as mock_commit_check, patch(
        "jnpr.junos.utils.config.Config.diff"
    ) as mock_diff, patch(
        "jnpr.junos.utils.config.Config.load"
    ) as mock_load, patch(
        "salt.utils.files.safe_rm"
    ) as mock_safe_rm, patch(
        "salt.utils.files.mkstemp"
    ) as mock_mkstemp, patch(
        "os.path.isfile"
    ) as mock_isfile, patch(
        "salt.utils.files.fopen", mock_open(), create=True
    ) as m_open, patch(
        "os.path.getsize"
    ) as mock_getsize:
        mock_isfile.return_value = True
        mock_getsize.return_value = 10
        mock_mkstemp.return_value = "test/path/config"
        mock_diff.return_value = "diff"
        mock_commit_check.return_value = True
        ret = {
            "message": "Commit check passed, but skipping commit for dry-run and rolling back configuration.",
            "out": True,
        }
        assert junos.install_config("actual/path/config", test=True) == ret
        mock_commit.assert_not_called()


def test_install_config_write_diff_dynamic_mode():
    with patch("jnpr.junos.utils.config.Config.commit") as mock_commit, patch(
        "jnpr.junos.utils.config.Config.commit_check"
    ) as mock_commit_check, patch(
        "jnpr.junos.utils.config.Config.diff"
    ) as mock_diff, patch(
        "jnpr.junos.utils.config.Config.load"
    ) as mock_load, patch(
        "salt.utils.files.safe_rm"
    ) as mock_safe_rm, patch(
        "salt.utils.files.mkstemp"
    ) as mock_mkstemp, patch(
        "os.path.isfile"
    ) as mock_isfile, patch(
        "salt.utils.files.fopen", mock_open(), create=True
    ) as m_open, patch(
        "os.path.getsize"
    ) as mock_getsize:
        mock_isfile.return_value = True
        mock_getsize.return_value = 10
        mock_mkstemp.return_value = "test/path/config"
        mock_diff.return_value = "diff"
        mock_commit_check.return_value = True
        ret = {
            "message": "Write diff is not supported with dynamic/ephemeral configuration mode",
            "out": False,
        }
        assert (
            junos.install_config(
                "actual/path/config", mode="dynamic", diffs_file="/path/to/dif"
            )
            == ret
        )
        mock_commit.assert_not_called()


def test_install_config_unknown_mode():
    with patch("jnpr.junos.utils.config.Config.commit") as mock_commit, patch(
        "jnpr.junos.utils.config.Config.commit_check"
    ) as mock_commit_check, patch(
        "jnpr.junos.utils.config.Config.diff"
    ) as mock_diff, patch(
        "jnpr.junos.utils.config.Config.load"
    ) as mock_load, patch(
        "salt.utils.files.safe_rm"
    ) as mock_safe_rm, patch(
        "salt.utils.files.mkstemp"
    ) as mock_mkstemp, patch(
        "os.path.isfile"
    ) as mock_isfile, patch(
        "salt.utils.files.fopen", mock_open(), create=True
    ) as m_open, patch(
        "os.path.getsize"
    ) as mock_getsize:
        mock_isfile.return_value = True
        mock_getsize.return_value = 10
        mock_mkstemp.return_value = "test/path/config"
        mock_diff.return_value = "diff"
        mock_commit_check.return_value = True
        ret = {
            "message": "install_config failed due to: unsupported action: abcdef",
            "out": False,
        }
        assert junos.install_config("actual/path/config", mode="abcdef") == ret
        mock_commit.assert_not_called()


def test_zeroize():
    with patch("jnpr.junos.device.Device.cli") as mock_cli:
        result = junos.zeroize()
        mock_cli.assert_called_once_with("request system zeroize")
        ret = {
            "message": "Completed zeroize and rebooted",
            "out": True,
        }
        assert result == ret


def test_zeroize_throw_exception():
    with patch("jnpr.junos.device.Device.cli") as mock_cli:
        mock_cli.side_effect = raise_exception
        ret = {
            "message": 'Could not zeroize due to : "Test exception"',
            "out": False,
        }
        assert junos.zeroize() == ret


def test_install_os_without_args():
    ret = {
        "message": "Please provide the salt path where the junos image is present.",
        "out": False,
    }
    assert junos.install_os() == ret


def test_install_os_cp_fails():
    with patch.dict(
        junos.__salt__,
        {
            "cp.is_cached": MagicMock(return_value="/pat/to/tmp/file"),
            "cp.hash_file": MagicMock(
                return_value={"hash_type": "sha256", "hsum": "a386e49c17"}
            ),
            "file.get_hash": MagicMock(return_value="xxxx"),
            "file.rmdir": MagicMock(return_value="True"),
        },
    ):
        with patch("jnpr.junos.utils.sw.SW.install") as mock_install, patch(
            "salt.utils.files.safe_rm"
        ) as mock_safe_rm, patch(
            "salt.utils.files.fopen", mock_open(), create=True
        ) as m_open, patch(
            "salt.utils.files.mkstemp"
        ) as mock_mkstemp, patch(
            "os.path.isfile"
        ) as mock_isfile, patch(
            "os.path.getsize"
        ) as mock_getsize:
            mock_getsize.return_value = 10
            mock_isfile.return_value = False
            mock_install.return_value = (
                False,
                "Invalid path. Please provide a valid image path",
            )
            ret = {
                "message": "Installation failed. Reason: Invalid path. Please provide a valid image path",
                "out": False,
            }
            assert junos.install_os("salt://image/path/") == ret


def test_install_os_image_cp_fails():
    with patch.dict(
        junos.__salt__, {"file.file_exists": MagicMock(return_value=False)}
    ):
        ret = {
            "message": "Invalid path. Please provide a valid image path",
            "out": False,
        }
        assert junos.install_os("/image/path/") == ret


def test_install_os():
    with patch.dict(
        junos.__salt__,
        {
            "cp.is_cached": MagicMock(return_value="test/path/config"),
            "cp.hash_file": MagicMock(
                return_value={"hash_type": "sha256", "hsum": "a386e49c17"}
            ),
            "file.get_hash": MagicMock(return_value="a386e49c17"),
        },
    ):
        with patch("jnpr.junos.utils.sw.SW.install") as mock_install, patch(
            "salt.utils.files.safe_rm"
        ) as mock_safe_rm, patch(
            "salt.utils.files.fopen", mock_open(), create=True
        ) as m_open, patch(
            "salt.utils.files.mkstemp"
        ) as mock_mkstemp, patch(
            "os.path.isfile"
        ) as mock_isfile, patch(
            "os.path.getsize"
        ) as mock_getsize:
            mock_getsize.return_value = 10
            mock_isfile.return_value = True
            mock_install.return_value = True, "installed"
            ret = {
                "message": "Installed the os.",
                "out": True,
            }
            assert junos.install_os("path") == ret


def test_install_os_failure():
    with patch("jnpr.junos.utils.sw.SW.install") as mock_install, patch(
        "salt.utils.files.safe_rm"
    ) as mock_safe_rm, patch(
        "salt.utils.files.fopen", mock_open(), create=True
    ) as m_open, patch(
        "salt.utils.files.mkstemp"
    ) as mock_mkstemp, patch(
        "os.path.isfile"
    ) as mock_isfile, patch(
        "os.path.getsize"
    ) as mock_getsize:
        mock_getsize.return_value = 10
        mock_isfile.return_value = True
        mock_install.return_value = False, "because we are testing failure"
        ret = {
            "message": "Installation failed. Reason: because we are testing failure",
            "out": False,
        }
        assert junos.install_os("path") == ret


def test_install_os_with_reboot_arg():
    with patch("jnpr.junos.utils.sw.SW.install") as mock_install, patch(
        "jnpr.junos.utils.sw.SW.reboot"
    ) as mock_reboot, patch("salt.utils.files.safe_rm") as mock_safe_rm, patch(
        "salt.utils.files.fopen", mock_open(), create=True
    ) as m_open, patch(
        "salt.utils.files.mkstemp"
    ) as mock_mkstemp, patch(
        "os.path.isfile"
    ) as mock_isfile, patch(
        "os.path.getsize"
    ) as mock_getsize:
        mock_getsize.return_value = 10
        mock_isfile.return_value = True
        mock_install.return_value = True, "installed"
        args = {
            "__pub_user": "root",
            "__pub_arg": [{"reboot": True}],
            "reboot": True,
            "__pub_fun": "junos.install_os",
            "__pub_jid": "20170222213858582619",
            "__pub_tgt": "mac_min",
            "__pub_tgt_type": "glob",
            "__pub_ret": "",
        }
        ret = {
            "message": "Successfully installed and rebooted!",
            "out": True,
        }
        assert junos.install_os("path", **args) == ret


def test_install_os_pyez_install_throws_exception():
    with patch("jnpr.junos.utils.sw.SW.install") as mock_install, patch(
        "salt.utils.files.safe_rm"
    ) as mock_safe_rm, patch(
        "salt.utils.files.fopen", mock_open(), create=True
    ) as m_open, patch(
        "salt.utils.files.mkstemp"
    ) as mock_mkstemp, patch(
        "os.path.isfile"
    ) as mock_isfile, patch(
        "os.path.getsize"
    ) as mock_getsize:
        mock_getsize.return_value = 10
        mock_isfile.return_value = True
        mock_install.side_effect = raise_exception
        ret = {
            "message": 'Installation failed due to: "Test exception"',
            "out": False,
        }
        assert junos.install_os("path") == ret


def test_install_os_with_reboot_raises_exception():
    with patch("jnpr.junos.utils.sw.SW.install") as mock_install, patch(
        "jnpr.junos.utils.sw.SW.reboot"
    ) as mock_reboot, patch("salt.utils.files.safe_rm") as mock_safe_rm, patch(
        "salt.utils.files.fopen", mock_open(), create=True
    ) as m_open, patch(
        "salt.utils.files.mkstemp"
    ) as mock_mkstemp, patch(
        "os.path.isfile"
    ) as mock_isfile, patch(
        "os.path.getsize"
    ) as mock_getsize:
        mock_getsize.return_value = 10
        mock_isfile.return_value = True
        mock_install.return_value = True, "installed"
        mock_reboot.side_effect = raise_exception
        args = {
            "__pub_user": "root",
            "__pub_arg": [{"reboot": True}],
            "reboot": True,
            "__pub_fun": "junos.install_os",
            "__pub_jid": "20170222213858582619",
            "__pub_tgt": "mac_min",
            "__pub_tgt_type": "glob",
            "__pub_ret": "",
        }
        ret = {
            "message": 'Installation successful but reboot failed due to : "Test exception"',
            "out": False,
        }
        assert junos.install_os("path", **args) == ret


def test_install_os_no_copy():
    with patch("jnpr.junos.utils.sw.SW.install") as mock_install, patch(
        "salt.utils.files.safe_rm"
    ) as mock_safe_rm, patch(
        "salt.utils.files.fopen", mock_open(), create=True
    ) as m_open, patch(
        "salt.utils.files.mkstemp"
    ) as mock_mkstemp, patch(
        "os.path.isfile"
    ) as mock_isfile, patch(
        "os.path.getsize"
    ) as mock_getsize:
        mock_getsize.return_value = 10
        mock_isfile.return_value = True
        mock_install.return_value = True, "installed"
        ret = {
            "message": "Installed the os.",
            "out": True,
        }
        assert junos.install_os("path", no_copy=True) == ret
        mock_install.assert_called_with(
            "path", no_copy=True, progress=True, timeout=1800
        )
        mock_mkstemp.assert_not_called()
        mock_safe_rm.assert_not_called()


def test_install_os_issu():
    with patch("jnpr.junos.utils.sw.SW.install") as mock_install, patch(
        "salt.utils.files.safe_rm"
    ) as mock_safe_rm, patch(
        "salt.utils.files.fopen", mock_open(), create=True
    ) as m_open, patch(
        "salt.utils.files.mkstemp"
    ) as mock_mkstemp, patch(
        "os.path.isfile"
    ) as mock_isfile, patch(
        "os.path.getsize"
    ) as mock_getsize:
        mock_getsize.return_value = 10
        mock_isfile.return_value = True
        mock_install.return_value = True, "installed"
        ret = {
            "message": "Installed the os.",
            "out": True,
        }
        assert junos.install_os("path", issu=True) == ret
        mock_install.assert_called_with(ANY, issu=True, progress=True, timeout=1800)


def test_install_os_add_params():
    with patch("jnpr.junos.utils.sw.SW.install") as mock_install, patch(
        "salt.utils.files.safe_rm"
    ) as mock_safe_rm, patch(
        "salt.utils.files.fopen", mock_open(), create=True
    ) as m_open, patch(
        "salt.utils.files.mkstemp"
    ) as mock_mkstemp, patch(
        "os.path.isfile"
    ) as mock_isfile, patch(
        "os.path.getsize"
    ) as mock_getsize:
        mock_getsize.return_value = 10
        mock_isfile.return_value = True
        mock_install.return_value = True, "installed"
        ret = {
            "message": "Installed the os.",
            "out": True,
        }
        remote_path = "/path/to/file"
        assert (
            junos.install_os("path", remote_path=remote_path, nssu=True, validate=True)
            == ret
        )
        mock_install.assert_called_with(
            ANY,
            nssu=True,
            remote_path=remote_path,
            progress=True,
            validate=True,
            timeout=1800,
        )


def test_file_copy_without_args():
    pytest.raises(TypeError, junos.file_copy)


@patch("paramiko.SSHClient")
@patch("scp.SCPClient.put")
@patch("scp.SCPClient.__init__")
def test_file_copy_invalid_src(mock_scpclient, mock_put, mock_ssh):
    mock_scpclient.return_value = None
    invalid_path = "invalid/file/path"
    mock_put.side_effect = Exception(invalid_path)
    with patch("os.path.isfile") as mock_isfile:
        mock_isfile.return_value = False
        ret = {
            "message": 'Could not copy file : "invalid/file/path"',
            "out": False,
        }
        assert junos.file_copy(invalid_path, "file") == ret


def test_file_copy_without_dest():
    pytest.raises(TypeError, junos.file_copy, src="/home/user/config.set")


def test_file_copy():
    with patch("salt.modules.junos.SCP") as mock_scp, patch(
        "os.path.isfile"
    ) as mock_isfile:
        mock_isfile.return_value = True
        ret = {
            "message": "Successfully copied file from test/src/file to file",
            "out": True,
        }
        assert junos.file_copy(dest="file", src="test/src/file") == ret


def test_file_copy_exception():
    with patch("salt.modules.junos.SCP") as mock_scp, patch(
        "os.path.isfile"
    ) as mock_isfile:
        mock_isfile.return_value = True
        mock_scp.side_effect = raise_exception
        ret = {
            "message": 'Could not copy file : "Test exception"',
            "out": False,
        }
        assert junos.file_copy(dest="file", src="test/src/file") == ret


# These test cases test the __virtual__ function, used internally by salt
# to check if the given module is loadable. This function is not used by
# an external user.


def test_virtual_proxy_unavailable():
    with patch.dict(junos.__opts__, {}):
        res = (
            False,
            "The junos or dependent module could not be loaded: "
            "junos-eznc or jxmlease or yamlordereddictloader or "
            "proxy could not be loaded.",
        )
        assert junos.__virtual__() == res


def test_virtual_all_true():
    with patch.dict(junos.__opts__, {"proxy": "test"}):
        assert junos.__virtual__() == "junos"


def test_rpc_without_args():
    ret = {
        "message": "Please provide the rpc to execute.",
        "out": False,
    }
    assert junos.rpc() == ret


def test_rpc_get_config_exception():
    with patch("jnpr.junos.device.Device.execute") as mock_execute:
        mock_execute.side_effect = raise_exception
        ret = {
            "message": 'RPC execution failed due to "Test exception"',
            "out": False,
        }
        assert junos.rpc("get_config") == ret


def test_rpc_get_config_filter():
    with patch("jnpr.junos.device.Device.execute") as mock_execute:
        mock_execute.return_value = etree.XML("<reply><rpc/></reply>")
        args = {
            "__pub_user": "root",
            "__pub_arg": [
                "get-config",
                {"filter": "<configuration><system/></configuration>"},
            ],
            "__pub_fun": "junos.rpc",
            "__pub_jid": "20170314162715866528",
            "__pub_tgt": "mac_min",
            "__pub_tgt_type": "glob",
            "filter": "<configuration><system/></configuration>",
            "__pub_ret": "",
        }
        junos.rpc("get-config", **args)
        exec_args = mock_execute.call_args
        expected_rpc = b'<get-configuration format="xml"><configuration><system/></configuration></get-configuration>'
        assert etree.tostring(exec_args[0][0]) == expected_rpc


def test_rpc_get_interface_information():
    with patch("jnpr.junos.device.Device.execute") as mock_execute:
        junos.rpc("get-interface-information", format="json")
        args = mock_execute.call_args
        expected_rpc = b'<get-interface-information format="json"/>'
        assert etree.tostring(args[0][0]) == expected_rpc


def test_rpc_get_interface_information_with_kwargs():
    with patch("jnpr.junos.device.Device.execute") as mock_execute:
        args = {
            "__pub_user": "root",
            "__pub_arg": [
                "get-interface-information",
                "",
                "text",
                {"terse": True, "interface_name": "lo0", "format": "text"},
            ],
            "format": "text",
            "terse": True,
            "__pub_fun": "junos.rpc",
            "__pub_jid": "20170314160943363563",
            "__pub_tgt": "mac_min",
            "interface_name": "lo0",
            "__pub_tgt_type": "glob",
            "__pub_ret": "",
        }
        junos.rpc("get-interface-information", **args)
        args = mock_execute.call_args
        expected_rpc = b'<get-interface-information format="text"><terse/><interface-name>lo0</interface-name></get-interface-information>'
        assert etree.tostring(args[0][0]) == expected_rpc


def test_rpc_get_chassis_inventory_filter_as_arg():
    with patch("salt.modules.junos.jxmlease.parse") as mock_jxmlease, patch(
        "salt.modules.junos.etree.tostring"
    ) as mock_tostring, patch(
        "salt.modules.junos.logging.Logger.warning"
    ) as mock_warning, patch(
        "jnpr.junos.device.Device.execute"
    ) as mock_execute:
        junos.rpc(
            "get-chassis-inventory",
            filter="<configuration><system/></configuration>",
        )
        mock_warning.assert_called_with(
            'Filter ignored as it is only used with "get-config" rpc'
        )


def test_rpc_get_interface_information_exception():
    with patch("jnpr.junos.device.Device.execute") as mock_execute:
        mock_execute.side_effect = raise_exception
        ret = {
            "message": 'RPC execution failed due to "Test exception"',
            "out": False,
        }
        assert junos.rpc("get_interface_information") == ret


def test_rpc_write_file_format_text():
    with patch("jnpr.junos.device.Device.execute") as mock_execute:
        mock_execute.return_value = etree.XML("<rpc-reply>text rpc reply</rpc-reply>")
        with patch("salt.utils.files.fopen", mock_open(), create=True) as m_open:
            junos.rpc("get-chassis-inventory", dest="/path/to/file", format="text")
            writes = m_open.write_calls()
            assert writes == ["text rpc reply"], writes


def test_rpc_write_file_format_json():
    with patch("jnpr.junos.device.Device.execute") as mock_execute, patch(
        "salt.utils.json.dumps"
    ) as mock_dumps:
        mock_dumps.return_value = "json rpc reply"
        with patch("salt.utils.files.fopen", mock_open(), create=True) as m_open:
            junos.rpc("get-chassis-inventory", dest="/path/to/file", format="json")
            writes = m_open.write_calls()
            assert writes == ["json rpc reply"], writes


def test_rpc_write_file():
    with patch("salt.modules.junos.jxmlease.parse") as mock_parse, patch(
        "salt.modules.junos.etree.tostring"
    ) as mock_tostring, patch("jnpr.junos.device.Device.execute") as mock_execute:
        mock_tostring.return_value = "xml rpc reply"
        with patch("salt.utils.files.fopen", mock_open(), create=True) as m_open:
            junos.rpc("get-chassis-inventory", dest="/path/to/file")
            writes = m_open.write_calls()
            assert writes == ["xml rpc reply"], writes


def test_lock_success():
    ret_exp = {"out": True, "message": "Successfully locked the configuration."}
    ret = junos.lock()
    assert ret == ret_exp


def test_lock_error():
    ret_exp = {"out": False, "message": 'Could not gain lock due to : "LockError"'}
    with patch("jnpr.junos.utils.config.Config.lock") as mock_lock:
        mock_lock.side_effect = LockError(None)
        ret = junos.lock()
        assert ret == ret_exp


def test_unlock_success():
    ret_exp = {"out": True, "message": "Successfully unlocked the configuration."}
    ret = junos.unlock()
    assert ret == ret_exp


def test_unlock_error():
    ret_exp = {
        "out": False,
        "message": 'Could not unlock configuration due to : "UnlockError"',
    }
    with patch("jnpr.junos.utils.config.Config.unlock") as mock_unlock:
        mock_unlock.side_effect = UnlockError(None)
        ret = junos.unlock()
        assert ret == ret_exp


def test_load_none_path():
    ret_exp = {
        "out": False,
        "message": ("Please provide the salt path where the configuration is present"),
    }
    ret = junos.load()
    assert ret == ret_exp


def test_load_wrong_tmp_file():
    ret_exp = {
        "out": False,
        "message": (
            'Could not load configuration due to : "[Errno 2] No such file or'
            " directory: '/pat/to/tmp/file'\""
        ),
        "format": "text",
    }
    with patch.dict(
        junos.__salt__,
        {
            "cp.is_cached": MagicMock(return_value="/pat/to/tmp/file"),
            "cp.hash_file": MagicMock(
                return_value={"hash_type": "sha256", "hsum": "a386e49c17"}
            ),
            "file.get_hash": MagicMock(return_value="a386e49c17"),
        },
    ):
        with patch("salt.utils.files.fopen", mock_open(), create=True) as m_open, patch(
            "os.path.getsize"
        ) as mock_getsize, patch("salt.utils.files.mkstemp") as mock_mkstmp:
            mock_mkstmp.return_value = "/pat/to/tmp/file"
            mock_getsize.return_value = 1000
            ret = junos.load("salt://path/to/file")
            assert ret == ret_exp


def test_load_invalid_path():
    with patch("salt.utils.files.mkstemp") as mock_mkstmp:
        mock_mkstmp.return_value = "/path/to/file"
        pytest.raises(FileNotFoundError, junos.load, path="/path/to/file")


def test_load_no_extension():
    ret_exp = {"out": True, "message": "Successfully loaded the configuration."}
    with patch("os.path.getsize") as mock_getsize, patch(
        "jnpr.junos.utils.config.Config.load"
    ) as mock_load, patch(
        "salt.utils.files.fopen", mock_open(), create=True
    ) as m_open, patch(
        "salt.utils.files.mkstemp"
    ) as mock_mkstmp, patch(
        "os.path.isfile"
    ) as mock_isfile:
        mock_getsize.return_value = 1000
        mock_mkstmp.return_value = "/path/to/file"
        mock_isfile.return_value = True
        ret = junos.load("/path/to/file")
        mock_load.assert_called_with(format="text", path="/path/to/file")
        assert ret == ret_exp


def test_load_xml_extension():
    ret_exp = {"out": True, "message": "Successfully loaded the configuration."}
    with patch("os.path.getsize") as mock_getsize, patch(
        "jnpr.junos.utils.config.Config.load"
    ) as mock_load, patch("os.path.isfile") as mock_isfile, patch(
        "salt.utils.files.fopen", mock_open(), create=True
    ) as m_open, patch(
        "salt.utils.files.mkstemp"
    ) as mock_mkstmp:
        mock_getsize.return_value = 1000
        mock_mkstmp.return_value = "/path/to/file.xml"
        mock_isfile.return_value = True
        ret = junos.load("/path/to/file.xml")
        mock_load.assert_called_with(format="xml", path="/path/to/file.xml")
        assert ret == ret_exp


def test_load_xml_extension_with_kwargs():
    ret_exp = {"out": True, "message": "Successfully loaded the configuration."}
    with patch("os.path.getsize") as mock_getsize, patch(
        "jnpr.junos.utils.config.Config.load"
    ) as mock_load, patch("salt.utils.files.mkstemp") as mock_mkstmp, patch(
        "os.path.isfile"
    ) as mock_isfile, patch(
        "salt.utils.files.fopen"
    ) as fopen, patch(
        "salt.utils.files.mkstemp"
    ) as mock_mkstmp:
        mock_mkstmp.return_value = "/path/to/file"
        mock_isfile.return_value = True
        ret = junos.load("/path/to/file.xml", template_vars=dict(hostname="test"))
        mock_load.assert_called_with(
            format="xml", path="/path/to/file", template_vars={"hostname": "test"}
        )
        assert ret == ret_exp


def test_load_set_extension():
    ret_exp = {"out": True, "message": "Successfully loaded the configuration."}
    with patch("os.path.getsize") as mock_getsize, patch(
        "jnpr.junos.utils.config.Config.load"
    ) as mock_load, patch("salt.utils.files.mkstemp") as mock_mkstmp, patch(
        "salt.utils.files.fopen", mock_open(), create=True
    ) as m_open, patch(
        "os.path.isfile"
    ) as mock_isfile:
        mock_getsize.return_value = 1000
        mock_mkstmp.return_value = "/path/to/file.set"
        mock_isfile.return_value = True
        ret = junos.load("/path/to/file.set")
        mock_load.assert_called_with(format="set", path="/path/to/file.set")
        assert ret == ret_exp


def test_load_replace_true():
    ret_exp = {"out": True, "message": "Successfully loaded the configuration."}
    with patch("os.path.getsize") as mock_getsize, patch(
        "jnpr.junos.utils.config.Config.load"
    ) as mock_load, patch("salt.utils.files.mkstemp") as mock_mkstmp, patch(
        "salt.utils.files.fopen", mock_open(), create=True
    ) as m_open, patch(
        "os.path.isfile"
    ) as mock_isfile:
        mock_getsize.return_value = 1000
        mock_mkstmp.return_value = "/path/to/file"
        mock_isfile.return_value = True
        ret = junos.load("/path/to/file", replace=True)
        mock_load.assert_called_with(format="text", merge=False, path="/path/to/file")
        assert ret == ret_exp


def test_load_replace_false():
    ret_exp = {"out": True, "message": "Successfully loaded the configuration."}
    with patch("os.path.getsize") as mock_getsize, patch(
        "jnpr.junos.utils.config.Config.load"
    ) as mock_load, patch("salt.utils.files.mkstemp") as mock_mkstmp, patch(
        "salt.utils.files.fopen", mock_open(), create=True
    ) as m_open, patch(
        "os.path.isfile"
    ) as mock_isfile:
        mock_getsize.return_value = 1000
        mock_mkstmp.return_value = "/path/to/file"
        mock_isfile.return_value = True
        ret = junos.load("/path/to/file", replace=False)
        mock_load.assert_called_with(format="text", replace=False, path="/path/to/file")
        assert ret == ret_exp


def test_load_overwrite_true():
    ret_exp = {"out": True, "message": "Successfully loaded the configuration."}
    with patch("os.path.getsize") as mock_getsize, patch(
        "jnpr.junos.utils.config.Config.load"
    ) as mock_load, patch("salt.utils.files.mkstemp") as mock_mkstmp, patch(
        "salt.utils.files.fopen", mock_open(), create=True
    ) as m_open, patch(
        "os.path.isfile"
    ) as mock_isfile:
        mock_getsize.return_value = 1000
        mock_mkstmp.return_value = "/path/to/file"
        mock_isfile.return_value = True
        ret = junos.load("/path/to/file", overwrite=True)
        mock_load.assert_called_with(
            format="text", overwrite=True, path="/path/to/file"
        )
        assert ret == ret_exp


def test_load_overwrite_false():
    ret_exp = {"out": True, "message": "Successfully loaded the configuration."}
    with patch("os.path.getsize") as mock_getsize, patch(
        "jnpr.junos.utils.config.Config.load"
    ) as mock_load, patch(
        "salt.utils.files.fopen", mock_open(), create=True
    ) as m_open, patch(
        "salt.utils.files.mkstemp"
    ) as mock_mkstmp, patch(
        "os.path.isfile"
    ) as mock_isfile:
        mock_getsize.return_value = 1000
        mock_mkstmp.return_value = "/path/to/file"
        mock_isfile.return_value = True
        ret = junos.load("/path/to/file", overwrite=False)
        mock_load.assert_called_with(format="text", merge=True, path="/path/to/file")
        assert ret == ret_exp


def test_load_error():
    ret_exp = {
        "out": False,
        "format": "text",
        "message": 'Could not load configuration due to : "Test Error"',
    }
    with patch("os.path.getsize") as mock_getsize, patch(
        "jnpr.junos.utils.config.Config.load"
    ) as mock_load, patch("salt.utils.files.mkstemp") as mock_mkstmp, patch(
        "salt.utils.files.fopen", mock_open(), create=True
    ) as m_open, patch(
        "os.path.isfile"
    ) as mock_isfile:
        mock_getsize.return_value = 1000
        mock_mkstmp.return_value = "/path/to/file"
        mock_isfile.return_value = True
        mock_load.side_effect = Exception("Test Error")
        ret = junos.load("/path/to/file")
        assert ret == ret_exp


def test_load_template():
    ret_exp = {
        "out": True,
        "message": "Successfully loaded the configuration.",
    }
    with patch("os.path.getsize") as mock_getsize, patch(
        "jnpr.junos.utils.config.Config.load"
    ) as mock_load:
        ret = junos.load("tests/unit/modules/templates/basic2.j2", test=True)
        assert ret == ret_exp


def test_commit_check_success():
    ret_exp = {"out": True, "message": "Commit check succeeded."}
    ret = junos.commit_check()
    assert ret == ret_exp


def test_commit_check_error():
    ret_exp = {"out": False, "message": "Commit check failed with "}
    with patch("jnpr.junos.utils.config.Config.commit_check") as mock_check:
        mock_check.side_effect = Exception
        ret = junos.commit_check()
        assert ret == ret_exp


def test_get_table_wrong_path():
    table = "ModuleTable"
    file = "sample.yml"
    path = "/path/to/file"
    ret_exp = {
        "out": False,
        "hostname": "1.1.1.1",
        "tablename": "ModuleTable",
        "message": f"Given table file {file} cannot be located",
    }
    with patch.dict(
        junos.__salt__, {"file.file_exists": MagicMock(return_value=False)}
    ):
        with patch("jnpr.junos.factory.FactoryLoader.load") as mock_load, patch(
            "salt.utils.files.fopen"
        ) as mock_fopen, patch("jnpr.junos.factory.FactoryLoader.load") as mock_load:
            ret = junos.get_table(table, file, path)
            assert ret == ret_exp
            mock_load.assert_not_called()


def test_get_table_no_path_no_file():
    table = "ModuleTable"
    file = "inventory.yml"
    ret_exp = {
        "out": False,
        "hostname": "1.1.1.1",
        "tablename": "ModuleTable",
        "message": f"Given table file {file} cannot be located",
    }
    with patch.dict(
        junos.__salt__, {"file.file_exists": MagicMock(return_value=False)}
    ):
        with patch("jnpr.junos.factory.FactoryLoader.load") as mock_load, patch(
            "glob.glob"
        ) as mock_fopen:
            mock_fopen.return_value = []
            ret = junos.get_table(table, file)
            assert ret == ret_exp
            mock_load.assert_not_called()


def test_get_table_yaml_load_error():
    table = "ModuleTable"
    file = "inventory.yml"
    path = "/path/to/file"
    message = "File not located test"
    ret_exp = {
        "out": False,
        "hostname": "1.1.1.1",
        "tablename": "ModuleTable",
        "message": "Uncaught exception during YAML Load - please report: {}".format(
            message
        ),
    }
    with patch("salt.utils.files.fopen", mock_open(), create=True) as mock_file, patch(
        "glob.glob"
    ) as mock_fopen, patch.object(yaml, "load") as mock_yamlload:
        mock_fopen.return_value = ["/path/to/file"]
        mock_yamlload.side_effect = OSError(message)
        ret = junos.get_table(table, file, path)
        assert ret == ret_exp


def test_get_table_api_error():
    table = "sample"
    file = "inventory.yml"
    table_yamlload = {
        "ModuleTable": {
            "item": (
                ".//chassis-sub-module|.//chassis-module|.//chassis-sub-sub-module"
            ),
            "key": "name",
            "rpc": "get-chassis-inventory",
            "view": "ModuleTableView",
        },
        "ModuleTableView": {
            "fields": {
                "jname": "name",
                "pn": "part-number",
                "sn": "serial-number",
                "type": "description",
                "ver": "version",
            },
        },
    }
    ret_exp = {
        "out": False,
        "hostname": "1.1.1.1",
        "tablename": "sample",
        "message": (
            "Uncaught exception during get API call - please report: '{}'".format(
                str(table)
            )
        ),
    }
    with patch("jnpr.junos.device.Device.execute") as mock_execute, patch(
        "yaml.load"
    ) as mock_yamlload, patch(
        "salt.utils.files.fopen", mock_open(), create=True
    ) as m_open:
        mock_yamlload.return_value = table_yamlload
        ret = junos.get_table(table, file)
        assert ret["out"] == ret_exp["out"]
        assert ret["tablename"] == ret_exp["tablename"]
        assert ret["message"] == ret_exp["message"]


def test_get_table_connect_closed_error():
    table = "ModuleTable"
    file = "inventory.yml"
    table_yamlload = {
        "ModuleTable": {
            "item": (
                ".//chassis-sub-module|.//chassis-module|.//chassis-sub-sub-module"
            ),
            "key": "name",
            "rpc": "get-chassis-inventory",
            "view": "ModuleTableView",
        },
        "ModuleTableView": {
            "fields": {
                "jname": "name",
                "pn": "part-number",
                "sn": "serial-number",
                "type": "description",
                "ver": "version",
            },
        },
    }
    ret_exp = {
        "out": False,
        "hostname": "1.1.1.1",
        "tablename": "ModuleTable",
        "message": (
            "Got ConnectClosedError exception. Connection lost with Device(1.1.1.1)"
        ),
    }
    with patch("jnpr.junos.factory.optable.OpTable.get") as mock_load, patch(
        "yaml.load"
    ) as mock_yamlload, patch(
        "salt.utils.files.fopen", mock_open(), create=True
    ) as m_open:
        dev = Device(host="1.1.1.1", user="rick")
        mock_load.side_effect = ConnectClosedError(dev)
        mock_yamlload.return_value = table_yamlload
        ret = junos.get_table(table, file)
        assert ret["out"] == ret_exp["out"]
        assert ret["tablename"] == ret_exp["tablename"]
        assert ret["message"] == ret_exp["message"]


def test_get_table_inventory():
    table = "ModuleTable"
    file = "inventory.yml"
    pyez_tables_path = os.path.dirname(os.path.abspath(tables_dir.__file__))
    path = pyez_tables_path
    table_yamlload = {
        "ModuleTable": {
            "item": (
                ".//chassis-sub-module|.//chassis-module|.//chassis-sub-sub-module"
            ),
            "key": "name",
            "rpc": "get-chassis-inventory",
            "view": "ModuleTableView",
        },
        "ModuleTableView": {
            "fields": {
                "jname": "name",
                "pn": "part-number",
                "sn": "serial-number",
                "type": "description",
                "ver": "version",
            },
        },
    }
    with patch("jnpr.junos.device.Device.execute") as mock_execute, patch(
        "salt.utils.files.fopen", mock_open(), create=True
    ) as m_open, patch("yaml.load") as mock_yamlload, patch(
        "salt.utils.json.dumps"
    ) as mock_dumps:
        mock_dumps.return_value = "json rpc reply"
        mock_yamlload.return_value = table_yamlload
        ret = junos.get_table(table, file, path)
        assert ret["out"]


def test_get_table_no_path_inventory():
    table = "ModuleTable"
    file = "inventory.yml"
    table_yamlload = {
        "ModuleTable": {
            "item": (
                ".//chassis-sub-module|.//chassis-module|.//chassis-sub-sub-module"
            ),
            "key": "name",
            "rpc": "get-chassis-inventory",
            "view": "ModuleTableView",
        },
        "ModuleTableView": {
            "fields": {
                "jname": "name",
                "pn": "part-number",
                "sn": "serial-number",
                "type": "description",
                "ver": "version",
            },
        },
    }
    with patch("jnpr.junos.device.Device.execute") as mock_execute, patch(
        "salt.utils.files.fopen", mock_open(), create=True
    ) as m_open, patch("yaml.load") as mock_yamlload, patch(
        "salt.utils.json.dumps"
    ) as mock_dumps:
        mock_dumps.return_value = "json rpc reply"
        mock_yamlload.return_value = table_yamlload
        ret = junos.get_table(table, file)
        assert ret["out"]



import urllib.parse

import pytest
from tornado.httpclient import HTTPError


@pytest.mark.netapi_client_data(["local"])
async def test_run_good_login(http_client, auth_creds):
    """
    Test the run URL with good auth credentials
    """
    low = {"client": "local", "tgt": "*", "fun": "test.ping", **auth_creds}
    body = urllib.parse.urlencode(low)

    response = await http_client.fetch("/run", method="POST", body=body)
    assert response.code == 200


async def test_run_netapi_client_not_set(http_client, auth_creds):
    """
    Test the run URL with good auth credentials
    """
    low = {"client": "local", "tgt": "*", "fun": "test.ping", **auth_creds}
    body = urllib.parse.urlencode(low)

    response = await http_client.fetch(
        "/run", method="POST", body=body, raise_error=False
    )
    assert response.code == 400
    assert (
        "Client disabled: 'local'. Add to 'netapi_enable_clients' master config option to enable"
        in response.body
    )


@pytest.mark.netapi_client_data(["local"])
async def test_run_bad_login(http_client):
    """
    Test the run URL with bad auth credentials
    """
    low = {
        "client": "local",
        "tgt": "*",
        "fun": "test.ping",
        **{"totally": "invalid_creds"},
    }
    body = urllib.parse.urlencode(low)
    with pytest.raises(HTTPError) as exc:
        await http_client.fetch(
            "/run",
            method="POST",
            body=body,
        )
    assert exc.value.code == 401


@pytest.mark.netapi_client_data(["local"])
async def test_run_empty_token(http_client):
    """
    Test the run URL with empty token
    """
    low = {"client": "local", "tgt": "*", "fun": "test.ping", **{"token": ""}}
    body = urllib.parse.urlencode(low)
    with pytest.raises(HTTPError) as exc:
        await http_client.fetch(
            "/run",
            method="POST",
            body=body,
        )
    assert exc.value.code == 401


@pytest.mark.netapi_client_data(["local"])
async def test_run_empty_token_upercase(http_client):
    """
    Test the run URL with empty token with upercase characters
    """
    low = {"client": "local", "tgt": "*", "fun": "test.ping", **{"ToKen": ""}}
    body = urllib.parse.urlencode(low)
    with pytest.raises(HTTPError) as exc:
        await http_client.fetch(
            "/run",
            method="POST",
            body=body,
        )
    assert exc.value.code == 401


@pytest.mark.netapi_client_data(["local"])
async def test_run_wrong_token(http_client):
    """
    Test the run URL with incorrect token
    """
    low = {"client": "local", "tgt": "*", "fun": "test.ping", **{"token": "bad"}}
    body = urllib.parse.urlencode(low)
    with pytest.raises(HTTPError) as exc:
        await http_client.fetch(
            "/run",
            method="POST",
            body=body,
        )
    assert exc.value.code == 401


@pytest.mark.netapi_client_data(["local"])
async def test_run_pathname_token(http_client):
    """
    Test the run URL with path that exists in token
    """
    low = {
        "client": "local",
        "tgt": "*",
        "fun": "test.ping",
        **{"token": "/etc/passwd"},
    }
    body = urllib.parse.urlencode(low)
    with pytest.raises(HTTPError) as exc:
        await http_client.fetch(
            "/run",
            method="POST",
            body=body,
        )
    assert exc.value.code == 401


@pytest.mark.netapi_client_data(["local"])
async def test_run_pathname_not_exists_token(http_client):
    """
    Test the run URL with path that does not exist in token
    """
    low = {
        "client": "local",
        "tgt": "*",
        "fun": "test.ping",
        **{"token": "/tmp/does-not-exist"},
    }
    body = urllib.parse.urlencode(low)
    with pytest.raises(HTTPError) as exc:
        await http_client.fetch(
            "/run",
            method="POST",
            body=body,
        )
    assert exc.value.code == 401


@pytest.mark.slow_test
@pytest.mark.netapi_client_data(["local"])
async def test_run_extra_parameters(http_client, auth_creds):
    """
    Test the run URL with good auth credentials
    """
    low = {"client": "local", "tgt": "*", "fun": "test.ping", **auth_creds}
    low["id_"] = "some-minion-name"
    body = urllib.parse.urlencode(low)

    response = await http_client.fetch("/run", method="POST", body=body)
    assert response.code == 200



"""
:codeauthor: Shane Lee <slee@saltstack.com>
"""

import pytest

import salt.modules.win_lgpo as win_lgpo

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.skip_unless_on_windows,
    pytest.mark.slow_test,
]


@pytest.fixture
def reg_pol_dword():
    data = (
        b"PReg\x01\x00\x00\x00"  # Header
        b"[\x00"  # Opening list of policies
        b"S\x00o\x00m\x00e\x00\\\x00K\x00e\x00y\x00\x00\x00;\x00"  # Key
        b"V\x00a\x00l\x00u\x00e\x00N\x00a\x00m\x00e\x00\x00\x00;\x00"  # Value
        b"\x04\x00\x00\x00;\x00"  # Reg DWord Type
        b"\x04\x00\x00\x00;\x00"  # Size
        # b"\x01\x00\x00\x00"  # Reg Dword Data
        b"\x00\x00\x00\x00"  # No Data
        b"]\x00"  # Closing list of policies
    )
    yield data


def test_get_data_from_reg_pol_data(reg_pol_dword):
    encoded_name = "ValueName".encode("utf-16-le")
    encoded_null = chr(0).encode("utf-16-le")
    encoded_semicolon = ";".encode("utf-16-le")
    encoded_type = chr(4).encode("utf-16-le")
    encoded_size = chr(4).encode("utf-16-le")
    search_string = b"".join(
        [
            encoded_semicolon,
            encoded_name,
            encoded_null,
            encoded_semicolon,
            encoded_type,
            encoded_null,
            encoded_semicolon,
            encoded_size,
            encoded_null,
        ]
    )
    result = win_lgpo._getDataFromRegPolData(
        search_string, reg_pol_dword, return_value_name=True
    )
    assert result == {"ValueName": 0}



import pytest

pytestmark = [pytest.mark.slow_test]


def test_items(salt_ssh_cli):
    ret = salt_ssh_cli.run("config.items")
    assert ret.returncode == 0
    assert isinstance(ret.data, dict)
    assert "id" in ret.data
    assert "grains" in ret.data
    assert "__master_opts__" in ret.data
    assert "cachedir" in ret.data


@pytest.mark.parametrize("omit", (False, True))
def test_option_minion_opt(salt_ssh_cli, omit):
    # Minion opt
    ret = salt_ssh_cli.run(
        "config.option", "id", omit_opts=omit, omit_grains=True, omit_master=True
    )
    assert ret.returncode == 0
    assert (ret.data != salt_ssh_cli.get_minion_tgt()) is omit
    assert (ret.data == "") is omit


@pytest.mark.parametrize("omit", (False, True))
def test_option_pillar(salt_ssh_cli, omit):
    ret = salt_ssh_cli.run("config.option", "ext_spam", omit_pillar=omit)
    assert ret.returncode == 0
    assert (ret.data != "eggs") is omit
    assert (ret.data == "") is omit


@pytest.mark.parametrize("omit", (False, True))
def test_option_grain(salt_ssh_cli, omit):
    ret = salt_ssh_cli.run("config.option", "kernel", omit_grains=omit)
    assert ret.returncode == 0
    assert (
        ret.data not in ("Darwin", "Linux", "FreeBSD", "OpenBSD", "Windows")
    ) is omit
    assert (ret.data == "") is omit


@pytest.mark.parametrize("omit", (False, True))
def test_get_minion_opt(salt_ssh_cli, omit):
    ret = salt_ssh_cli.run("config.get", "cachedir", omit_master=True, omit_opts=omit)
    assert ret.returncode == 0
    assert (ret.data == "") is omit
    assert ("minion" not in ret.data) is omit


@pytest.mark.parametrize("omit", (False, True))
def test_get_pillar(salt_ssh_cli, omit):
    ret = salt_ssh_cli.run("config.get", "ext_spam", omit_pillar=omit)
    assert ret.returncode == 0
    assert (ret.data != "eggs") is omit
    assert (ret.data == "") is omit


@pytest.mark.parametrize("omit", (False, True))
def test_get_grain(salt_ssh_cli, omit):
    ret = salt_ssh_cli.run("config.get", "kernel", omit_grains=omit)
    assert ret.returncode == 0
    assert (
        ret.data not in ("Darwin", "Linux", "FreeBSD", "OpenBSD", "Windows")
    ) is omit
    assert (ret.data == "") is omit



"""
    Test cases for salt.modules.xfs
"""

import textwrap

import pytest

import salt.modules.xfs as xfs

pytestmark = [
    pytest.mark.skip_on_windows(reason="xfs not available on windows"),
    pytest.mark.skip_on_darwin(reason="xfs not available on darwin."),
]


@pytest.fixture
def configure_loader_modules():
    return {xfs: {}}


def test_blkid_output():
    """
    Test xfs._blkid_output when there is data
    """
    blkid_export = textwrap.dedent(
        """
        DEVNAME=/dev/sda1
        UUID=XXXXXXXX-XXXX-XXXX-XXXX-XXXXXXXXXXXX
        TYPE=xfs
        PARTUUID=YYYYYYYY-YY

        DEVNAME=/dev/sdb1
        PARTUUID=ZZZZZZZZ-ZZZZ-ZZZZ-ZZZZ-ZZZZZZZZZZZZ
        """
    )
    # We expect to find only data from /dev/sda1, nothig from
    # /dev/sdb1
    assert xfs._blkid_output(blkid_export) == {
        "/dev/sda1": {
            "label": None,
            "partuuid": "YYYYYYYY-YY",
            "uuid": "XXXXXXXX-XXXX-XXXX-XXXX-XXXXXXXXXXXX",
        }
    }


def test_parse_xfs_info():
    """
    Test parsing output from mkfs.xfs.
    """
    data = textwrap.dedent(
        """
        meta-data=/dev/vg00/testvol      isize=512    agcount=4, agsize=1310720 blks
                 =                       sectsz=4096  attr=2, projid32bit=1
                 =                       crc=1        finobt=1, sparse=1, rmapbt=0
                 =                       reflink=1
        data     =                       bsize=4096   blocks=5242880, imaxpct=25
                 =                       sunit=0      swidth=0 blks
        naming   =version 2              bsize=4096   ascii-ci=0, ftype=1
        log      =internal log           bsize=4096   blocks=2560, version=2
                 =                       sectsz=4096  sunit=1 blks, lazy-count=1
        realtime =none                   extsz=4096   blocks=0, rtextents=0
        Discarding blocks...Done.
        """
    )

    assert xfs._parse_xfs_info(data) == {
        "meta-data": {
            "section": "/dev/vg00/testvol",
            "isize": "512",
            "agcount": "4",
            "agsize": "1310720 blks",
            "sectsz": "4096",
            "attr": "2",
            "projid32bit": "1",
            "crc": "1",
            "finobt": "1",
            "sparse": "1",
            "rmapbt": "0",
            "reflink": "1",
        },
        "data": {
            "section": "data",
            "bsize": "4096",
            "blocks": "5242880",
            "imaxpct": "25",
            "sunit": "0",
            "swidth": "0 blks",
        },
        "naming": {
            "section": "version 2",
            "bsize": "4096",
            "ascii-ci": "0",
            "ftype": "1",
        },
        "log": {
            "section": "internal log",
            "bsize": "4096",
            "blocks": "2560",
            "version": "2",
            "sectsz": "4096",
            "sunit": "1 blks",
            "lazy-count": "1",
        },
        "realtime": {
            "section": "none",
            "extsz": "4096",
            "blocks": "0",
            "rtextents": "0",
        },
    }



"""
:codeauthor: Shane Lee <slee@saltstack.com>
"""

import pytest

import salt.config
import salt.grains.core
import salt.loader
import salt.modules.cmdmod as cmdmod
import salt.modules.win_file as win_file
import salt.modules.win_lgpo as win_lgpo
import salt.states.win_lgpo
import salt.utils.files
import salt.utils.platform
import salt.utils.stringutils
import salt.utils.win_dacl as win_dacl
import salt.utils.win_lgpo_auditpol as win_lgpo_auditpol
import salt.utils.win_reg as win_reg

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.skip_unless_on_windows,
    pytest.mark.slow_test,
]


@pytest.fixture
def configure_loader_modules(tmp_path):
    cachedir = tmp_path / "__test_admx_policy_cache_dir"
    cachedir.mkdir(parents=True, exist_ok=True)
    return {
        win_lgpo: {
            "__salt__": {
                "cmd.run": cmdmod.run,
                "file.file_exists": win_file.file_exists,
                "file.makedirs": win_file.makedirs_,
                "file.remove": win_file.remove,
                "file.write": win_file.write,
            },
            "__opts__": {
                "cachedir": str(cachedir),
            },
            "__utils__": {
                "auditpol.get_auditpol_dump": win_lgpo_auditpol.get_auditpol_dump,
                "reg.read_value": win_reg.read_value,
            },
        },
        win_file: {
            "__utils__": {
                "dacl.set_perms": win_dacl.set_perms,
            },
        },
    }


def _get_policy_adm_setting(
    policy_name, policy_class, return_full_policy_names, hierarchical_return
):
    """
    Helper function to get current setting
    """
    # Get the policy
    success, policy_obj, _, _ = salt.modules.win_lgpo._lookup_admin_template(
        policy_name=policy_name, policy_class=policy_class, adml_language="en-US"
    )
    if success:
        return salt.modules.win_lgpo._get_policy_adm_setting(
            admx_policy=policy_obj,
            policy_class=policy_class,
            adml_language="en-US",
            return_full_policy_names=return_full_policy_names,
            hierarchical_return=hierarchical_return,
        )
    return "Policy Not Found"


def test_not_configured():
    result = _get_policy_adm_setting(
        policy_name="Point and Print Restrictions",
        policy_class="Machine",
        return_full_policy_names=False,
        hierarchical_return=False,
    )
    expected = {"PointAndPrint_Restrictions_Win7": "Not Configured"}
    assert result == expected


def test_not_configured_hierarchical():
    result = _get_policy_adm_setting(
        policy_name="Point and Print Restrictions",
        policy_class="Machine",
        return_full_policy_names=False,
        hierarchical_return=True,
    )
    expected = {
        "Computer Configuration": {
            "Administrative Templates": {
                "Printers": {"PointAndPrint_Restrictions_Win7": "Not Configured"}
            }
        }
    }
    assert result == expected


def test_not_configured_full_names():
    result = _get_policy_adm_setting(
        policy_name="Point and Print Restrictions",
        policy_class="Machine",
        return_full_policy_names=True,
        hierarchical_return=False,
    )
    expected = {"Printers\\Point and Print Restrictions": "Not Configured"}
    assert result == expected


def test_not_configured_full_names_hierarchical():
    result = _get_policy_adm_setting(
        policy_name="Point and Print Restrictions",
        policy_class="Machine",
        return_full_policy_names=True,
        hierarchical_return=True,
    )
    expected = {
        "Computer Configuration": {
            "Administrative Templates": {
                "Printers": {"Point and Print Restrictions": "Not Configured"}
            }
        }
    }
    assert result == expected



import pytest

import salt.loader
import salt.loader.lazy
import salt.modules.boto_vpc
import salt.modules.virt


@pytest.fixture
def minion_mods(minion_opts):
    utils = salt.loader.utils(minion_opts)
    return salt.loader.minion_mods(minion_opts, utils=utils)


@pytest.mark.skipif(
    not salt.modules.boto_vpc.HAS_BOTO, reason="boto must be installed."
)
def test_load_boto_vpc(minion_mods):
    func = None
    try:
        func = minion_mods["boto_vpc.check_vpc"]
    except KeyError:
        pytest.fail("loader should not raise KeyError")
    assert func is not None
    assert isinstance(func, salt.loader.lazy.LoadedFunc)


@pytest.mark.skipif(
    not salt.modules.virt.HAS_LIBVIRT, reason="libvirt-python must be installed."
)
def test_load_virt(minion_mods):
    func = None
    try:
        func = minion_mods["virt.ctrl_alt_del"]
    except KeyError:
        pytest.fail("loader should not raise KeyError")
    assert func is not None
    assert isinstance(func, salt.loader.lazy.LoadedFunc)



"""
    :codeauthor: Nicole Thomas <nicole@saltstack.com>
"""

import os
import random

import pytest

import salt.utils.files
from salt.exceptions import CommandExecutionError

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.destructive_test,
    pytest.mark.skip_if_not_root,
    pytest.mark.skip_unless_on_darwin,
]


@pytest.fixture(scope="function")
def assign_cmd():
    return "net.inet.icmp.timestamp"


@pytest.fixture(scope="function")
def config_file():
    return "/etc/sysctl.conf"


@pytest.fixture(scope="function")
def setup_teardown_vars(salt_call_cli, assign_cmd, config_file):
    has_conf = False
    ret = salt_call_cli.run("sysctl.get", assign_cmd, config_file)
    val = ret.data

    if val is None:
        pytest.skip(f"The call 'sysctl.get {assign_cmd}' returned: None")

    # If sysctl file is present, make a copy
    # Remove original file so we can replace it with test files
    if os.path.isfile(config_file):
        has_conf = True
        try:
            temp_sysctl_config = __copy_sysctl(config_file)
        except CommandExecutionError:
            msg = "Could not copy file: {0}"
            raise CommandExecutionError(msg.format(config_file))
        os.remove(config_file)

    try:
        yield val
    finally:
        ret = salt_call_cli.run("sysctl.get", assign_cmd)
        if ret.data != val:
            salt_call_cli.run("sysctl.assign", assign_cmd, val)

        if has_conf is True:
            # restore original sysctl file
            __restore_sysctl(config_file, temp_sysctl_config)

        if has_conf is False and os.path.isfile(config_file):
            # remove sysctl.conf created by tests
            os.remove(temp_sysctl_config)


def test_assign(salt_call_cli, assign_cmd, setup_teardown_vars):
    """
    Tests assigning a single sysctl parameter
    """
    val = setup_teardown_vars[0]

    try:
        rand = random.randint(0, 500)
        while rand == val:
            rand = random.randint(0, 500)
        salt_call_cli.run("sysctl.assign", assign_cmd, rand)
        ret = int(salt_call_cli.run("sysctl.get", assign_cmd))
        info = int(ret.data)
        try:
            assert rand == info
        except AssertionError:
            salt_call_cli.run("sysctl.assign", assign_cmd, val)
            raise
    except CommandExecutionError:
        salt_call_cli.run("sysctl.assign", assign_cmd, val)
        raise


def test_persist_new_file(salt_call_cli, assign_cmd, config_file):
    """
    Tests assigning a sysctl value to a system without a sysctl.conf file
    """
    # Always start with a clean/known sysctl.conf state
    if os.path.isfile(config_file):
        os.remove(config_file)
    try:
        salt_call_cli.run("sysctl.persist", assign_cmd, 10)
        line = f"{assign_cmd}={10}"
        found = __check_string(config_file, line)
        assert found
    except CommandExecutionError:
        os.remove(config_file)
        raise


def test_persist_already_set(salt_call_cli, config_file, setup_teardown_vars):
    """
    Tests assigning a sysctl value that is already set in sysctl.conf file
    """
    # Always start with a clean/known sysctl.conf state
    if os.path.isfile(config_file):
        os.remove(config_file)
    try:
        salt_call_cli.run("sysctl.persist", assign_cmd, 50)
        ret = salt_call_cli.run("sysctl.persist", assign_cmd, 50)
        assert ret.data == "Already set"
    except CommandExecutionError:
        os.remove(config_file)
        raise


def test_persist_apply_change(
    salt_call_cli, assign_cmd, config_file, setup_teardown_vars
):
    """
    Tests assigning a sysctl value and applying the change to system
    """
    val = setup_teardown_vars[0]

    # Always start with a clean/known sysctl.conf state
    if os.path.isfile(config_file):
        os.remove(config_file)
    try:
        rand = random.randint(0, 500)
        while rand == val:
            rand = random.randint(0, 500)
        salt_call_cli.run("sysctl.persist", assign_cmd, rand, apply_change=True)
        ret = salt_call_cli.run("sysctl.get", assign_cmd)
        info = int(ret.data)
        assert info == rand
    except CommandExecutionError:
        os.remove(config_file)
        raise


def __copy_sysctl(CONFIG):
    """
    Copies an existing sysconf file and returns temp file path. Copied
    file will be restored in tearDown
    """
    # Create new temporary file path and open needed files
    temp_path = salt.utils.files.mkstemp()
    with salt.utils.files.fopen(CONFIG, "r") as org_conf:
        with salt.utils.files.fopen(temp_path, "w") as temp_sysconf:
            # write sysctl lines to temp file
            for line in org_conf:
                temp_sysconf.write(line)
    return temp_path


def __restore_sysctl(sysctl_config, temp_sysctl_config):
    """
    Restores the original sysctl.conf file from temporary copy
    """
    # If sysctl testing file exists, delete it
    if os.path.isfile(sysctl_config):
        os.remove(sysctl_config)

    # write temp lines to sysctl file to restore
    with salt.utils.files.fopen(temp_sysctl_config, "r") as temp_sysctl:
        with salt.utils.files.fopen(sysctl_config, "w") as sysctl:
            for line in temp_sysctl:
                sysctl.write(line)

    # delete temporary file
    os.remove(temp_sysctl_config)


def __check_string(conf_file, to_find):
    """
    Returns True if given line is present in file
    """
    with salt.utils.files.fopen(conf_file, "r") as f_in:
        for line in f_in:
            if to_find in salt.utils.stringutils.to_unicode(line):
                return True
        return False



"""
Module for rebootmgr
:maintainer:    Alberto Planas <aplanas@suse.com>
:maturity:      new
:depends:       None
:platform:      Linux

.. versionadded:: 3004
"""

import logging
import re

import salt.exceptions

log = logging.getLogger(__name__)


def __virtual__():
    """rebootmgrctl command is required."""
    if __utils__["path.which"]("rebootmgrctl") is not None:
        return True
    else:
        return (False, "Module rebootmgt requires the command rebootmgrctl")


def _cmd(cmd, retcode=False):
    """Utility function to run commands."""
    result = __salt__["cmd.run_all"](cmd)
    if retcode:
        return result["retcode"]

    if result["retcode"]:
        raise salt.exceptions.CommandExecutionError(result["stderr"])

    return result["stdout"]


def version():
    """Return the version of rebootmgrd

    CLI Example:

    .. code-block:: bash

        salt microos rebootmgr version

    """
    cmd = ["rebootmgrctl", "--version"]

    return _cmd(cmd).split()[-1]


def is_active():
    """Check if the rebootmgrd is running and active or not.

    CLI Example:

    .. code-block:: bash

        salt microos rebootmgr is_active

    """
    cmd = ["rebootmgrctl", "is_active", "--quiet"]

    return _cmd(cmd, retcode=True) == 0


def reboot(order=None):
    """Tells rebootmgr to schedule a reboot.

    With the [now] option, a forced reboot is done, no lock from etcd
    is requested and a set maintenance window is ignored. With the
    [fast] option, a lock from etcd is requested if needed, but a
    defined maintenance window is ignored.

    order
        If specified, can be "now" or "fast"

    CLI Example:

    .. code-block:: bash

        salt microos rebootmgr reboot
        salt microos rebootmgt reboot order=now

    """
    if order and order not in ("now", "fast"):
        raise salt.exceptions.CommandExecutionError(
            "Order parameter, if specified, must be 'now' or 'fast'"
        )

    cmd = ["rebootmgrctl", "reboot"]
    if order:
        cmd.append(order)

    return _cmd(cmd)


def cancel():
    """Cancels an already running reboot.

    CLI Example:

    .. code-block:: bash

        salt microos rebootmgr cancel

    """
    cmd = ["rebootmgrctl", "cancel"]

    return _cmd(cmd)


def status():
    """Returns the current status of rebootmgrd.

    Valid returned values are:
      0 - No reboot requested
      1 - Reboot requested
      2 - Reboot requested, waiting for maintenance window
      3 - Reboot requested, waiting for etcd lock.

    CLI Example:

    .. code-block:: bash

        salt microos rebootmgr status

    """
    cmd = ["rebootmgrctl", "status", "--quiet"]

    return _cmd(cmd, retcode=True)


def set_strategy(strategy=None):
    """A new strategy to reboot the machine is set and written into
    /etc/rebootmgr.conf.

    strategy
        If specified, must be one of those options:

        best-effort - This is the default strategy. If etcd is
            running, etcd-lock is used. If no etcd is running, but a
            maintenance window is specified, the strategy will be
            maint-window. If no maintenance window is specified, the
            machine is immediately rebooted (instantly).

        etcd-lock - A lock at etcd for the specified lock-group will
            be acquired before reboot. If a maintenance window is
            specified, the lock is only acquired during this window.

        maint-window - Reboot does happen only during a specified
            maintenance window. If no window is specified, the
            instantly strategy is followed.

        instantly - Other services will be informed that a reboot will
            happen. Reboot will be done without getting any locks or
            waiting for a maintenance window.

        off - Reboot requests are temporary
            ignored. /etc/rebootmgr.conf is not modified.

    CLI Example:

    .. code-block:: bash

        salt microos rebootmgr set_strategy stragegy=off

    """
    if strategy and strategy not in (
        "best-effort",
        "etcd-lock",
        "maint-window",
        "instantly",
        "off",
    ):
        raise salt.exceptions.CommandExecutionError("Strategy parameter not valid")

    cmd = ["rebootmgrctl", "set-strategy"]
    if strategy:
        cmd.append(strategy)

    return _cmd(cmd)


def get_strategy():
    """The currently used reboot strategy of rebootmgrd will be printed.

    CLI Example:

    .. code-block:: bash

        salt microos rebootmgr get_strategy

    """
    cmd = ["rebootmgrctl", "get-strategy"]

    return _cmd(cmd).split(":")[-1].strip()


def set_window(time, duration):
    """Set's the maintenance window.

    time
        The format of time is the same as described in
        systemd.time(7).

    duration
         The format of duration is "[XXh][YYm]".

    CLI Example:

    .. code-block:: bash

        salt microos rebootmgr set_window time="Thu,Fri 2020-*-1,5 11:12:13" duration=1h

    """
    cmd = ["rebootmgrctl", "set-window", time, duration]

    return _cmd(cmd)


def get_window():
    """The currently set maintenance window will be printed.

    CLI Example:

    .. code-block:: bash

        salt microos rebootmgr get_window

    """
    cmd = ["rebootmgrctl", "get-window"]
    window = _cmd(cmd)

    return dict(
        zip(
            ("time", "duration"),
            re.search(
                r"Maintenance window is set to (.*), lasting (.*).", window
            ).groups(),
        )
    )


def set_group(group):
    """Set the group, to which this machine belongs to get a reboot lock
       from etcd.

    group
        Group name

    CLI Example:

    .. code-block:: bash

        salt microos rebootmgr set_group group=group_1

    """
    cmd = ["rebootmgrctl", "set-group", group]

    return _cmd(cmd)


def get_group():
    """The currently set lock group for etcd.

    CLI Example:

    .. code-block:: bash

        salt microos rebootmgr get_group

    """
    cmd = ["rebootmgrctl", "get-group"]
    group = _cmd(cmd)

    return re.search(r"Etcd lock group is set to (.*)", group).groups()[0]


def set_max(max_locks, group=None):
    """Set the maximal number of hosts in a group, which are allowed to
       reboot at the same time.

    number
        Maximal number of hosts in a group

    group
        Group name

    CLI Example:

    .. code-block:: bash

        salt microos rebootmgr set_max 4

    """
    cmd = ["rebootmgrctl", "set-max"]
    if group:
        cmd.extend(["--group", group])
    cmd.append(max_locks)

    return _cmd(cmd)


def lock(machine_id=None, group=None):
    """Lock a machine. If no group is specified, the local default group
       will be used. If no machine-id is specified, the local machine
       will be locked.

    machine_id
        The machine-id is a network wide, unique ID. Per default the
        ID from /etc/machine-id is used.

    group
        Group name

    CLI Example:

    .. code-block:: bash

        salt microos rebootmgr lock group=group1

    """
    cmd = ["rebootmgrctl", "lock"]
    if group:
        cmd.extend(["--group", group])
    if machine_id:
        cmd.append(machine_id)

    return _cmd(cmd)


def unlock(machine_id=None, group=None):
    """Unlock a machine. If no group is specified, the local default group
       will be used. If no machine-id is specified, the local machine
       will be locked.

    machine_id
        The machine-id is a network wide, unique ID. Per default the
        ID from /etc/machine-id is used.

    group
        Group name

    CLI Example:

    .. code-block:: bash

        salt microos rebootmgr unlock group=group1

    """
    cmd = ["rebootmgrctl", "unlock"]
    if group:
        cmd.extend(["--group", group])
    if machine_id:
        cmd.append(machine_id)

    return _cmd(cmd)



"""
Tests for the Keystone states
"""

import logging

import pytest

from tests.support.case import ModuleCase
from tests.support.mixins import SaltReturnAssertsMixin

log = logging.getLogger(__name__)

NO_KEYSTONE = False
try:
    import keystoneclient  # pylint: disable=import-error,unused-import
except ImportError:
    NO_KEYSTONE = True


@pytest.mark.skipif(
    NO_KEYSTONE,
    reason="Please install keystoneclient and a keystone server before running "
    "keystone integration tests.",
)
class KeystoneStateTest(ModuleCase, SaltReturnAssertsMixin):
    """
    Validate the keystone state
    """

    endpoint = "http://localhost:35357/v2.0"
    token = "administrator"

    @pytest.mark.destructive_test
    def setUp(self):
        ret = self.run_state(
            "keystone.service_present",
            name="keystone",
            description="OpenStack Identity",
            service_type="identity",
            connection_endpoint=self.endpoint,
            connection_token=self.token,
        )
        self.assertTrue(
            ret["keystone_|-keystone_|-keystone_|-service_present"]["result"]
        )

        ret = self.run_state(
            "keystone.endpoint_present",
            name="keystone",
            region="RegionOne",
            publicurl="http://localhost:5000/v2.0",
            internalurl="http://localhost:5000/v2.0",
            adminurl="http://localhost:35357/v2.0",
            connection_endpoint=self.endpoint,
            connection_token=self.token,
        )
        self.assertTrue(
            ret["keystone_|-keystone_|-keystone_|-endpoint_present"]["result"]
        )

        ret = self.run_state(
            "keystone.tenant_present",
            name="admin",
            description="Admin Project",
            connection_endpoint=self.endpoint,
            connection_token=self.token,
        )
        self.assertTrue(ret["keystone_|-admin_|-admin_|-tenant_present"]["result"])

        ret = self.run_state(
            "keystone.tenant_present",
            name="demo",
            description="Demo Project",
            connection_endpoint=self.endpoint,
            connection_token=self.token,
        )
        self.assertTrue(ret["keystone_|-demo_|-demo_|-tenant_present"]["result"])

        ret = self.run_state(
            "keystone.role_present",
            name="admin",
            connection_endpoint=self.endpoint,
            connection_token=self.token,
        )
        self.assertTrue(ret["keystone_|-admin_|-admin_|-role_present"]["result"])

        ret = self.run_state(
            "keystone.role_present",
            name="user",
            connection_endpoint=self.endpoint,
            connection_token=self.token,
        )
        self.assertTrue(ret["keystone_|-user_|-user_|-role_present"]["result"])

        ret = self.run_state(
            "keystone.user_present",
            name="admin",
            email="admin@example.com",
            password="adminpass",
            tenant="admin",
            roles={"admin": ["admin"]},
            connection_endpoint=self.endpoint,
            connection_token=self.token,
        )
        self.assertTrue(ret["keystone_|-admin_|-admin_|-user_present"]["result"])

        ret = self.run_state(
            "keystone.user_present",
            name="demo",
            email="demo@example.com",
            password="demopass",
            tenant="demo",
            roles={"demo": ["user"]},
            connection_endpoint=self.endpoint,
            connection_token=self.token,
        )
        self.assertTrue(ret["keystone_|-demo_|-demo_|-user_present"]["result"])

    @pytest.mark.destructive_test
    def test_keystone_v2(self):
        ret = self.run_state(
            "keystone.service_present",
            name="testv2",
            description="Nova Service",
            service_type="compute",
            profile="adminv2",
        )
        self.assertTrue(ret["keystone_|-testv2_|-testv2_|-service_present"]["result"])

        ret = self.run_state(
            "keystone.endpoint_present",
            name="nova",
            description="Nova Service",
            publicurl="http://localhost:8774/v2.1/%(tenant_id)s",
            internalurl="http://localhost:8774/v2.1/%(tenant_id)s",
            adminurl="http://localhost:8774/v2.1/%(tenant_id)s",
            region="RegionOne",
            profile="adminv2",
        )
        self.assertTrue(ret["keystone_|-nova_|-nova_|-endpoint_present"]["result"])

        # Region Two
        ret = self.run_state(
            "keystone.endpoint_present",
            name="nova",
            description="Nova Service",
            publicurl="http://localhost:8774/v2.1/%(tenant_id)s",
            internalurl="http://localhost:8774/v2.1/%(tenant_id)s",
            adminurl="http://localhost:8774/v2.1/%(tenant_id)s",
            region="RegionTwo",
            profile="adminv2",
        )
        self.assertTrue(ret["keystone_|-nova_|-nova_|-endpoint_present"]["result"])

        # Region One, change publicurl
        ret = self.run_state(
            "keystone.endpoint_present",
            name="nova",
            description="Nova Service",
            publicurl="http://127.0.0.1:8774/v2.1/%(tenant_id)s",
            internalurl="http://localhost:8774/v2.1/%(tenant_id)s",
            adminurl="http://localhost:8774/v2.1/%(tenant_id)s",
            region="RegionOne",
            profile="adminv2",
        )
        self.assertTrue(ret["keystone_|-nova_|-nova_|-endpoint_present"]["result"])

        ret = self.run_state(
            "keystone.endpoint_get", name="nova", region="RegionOne", profile="adminv2"
        )
        self.assertTrue(ret["keystone_|-testv3_|-testv3_|-endpoint_present"]["result"])
        self.assertTrue(
            ret["keystone_|-testv3_|-testv3_|-endpoint_present"][
                "publicurl"
            ].start_with("http://127.0.0.1")
        )

        ret = self.run_state(
            "keystone.tenant_present",
            name="test",
            description="Test Tenant",
            profile="adminv2",
        )
        self.assertTrue(ret["keystone_|-test_|-test_|-tenant_present"]["result"])

        ret = self.run_state("keystone.role_present", name="user", profile="adminv2")
        self.assertTrue(ret["keystone_|-user_|-user_|-role_present"]["result"])

        ret = self.run_state(
            "keystone.user_present",
            name="test",
            email="test@example.com",
            tenant="test",
            password="testpass",
            roles={"test": ["user"]},
            profile="adminv2",
        )
        self.assertTrue(ret["keystone_|-test_|-test_|-user_present"]["result"])

        ret = self.run_state(
            "keystone.service_absent", name="testv2", profile="adminv2"
        )
        self.assertTrue(ret["keystone_|-testv2_|-testv2_|-service_absent"]["result"])

    @pytest.mark.destructive_test
    def test_keystone_v3(self):
        ret = self.run_state(
            "keystone.service_present",
            name="testv3",
            description="Image Service",
            service_type="image",
            profile="adminv3",
        )
        self.assertTrue(ret["keystone_|-testv3_|-testv3_|-service_present"]["result"])

        ret = self.run_state(
            "keystone.endpoint_present",
            name="testv3",
            description="Glance Service",
            interface="public",
            url="http://localhost:9292",
            region="RegionOne",
            profile="adminv3",
        )
        self.assertTrue(ret["keystone_|-testv3_|-testv3_|-endpoint_present"]["result"])

        ret = self.run_state(
            "keystone.endpoint_present",
            name="testv3",
            description="Glance Service",
            interface="internal",
            url="http://localhost:9292",
            region="RegionOne",
            profile="adminv3",
        )
        self.assertTrue(ret["keystone_|-testv3_|-testv3_|-endpoint_present"]["result"])

        ret = self.run_state(
            "keystone.endpoint_present",
            name="testv3",
            description="Glance Service",
            interface="admin",
            url="http://localhost:9292",
            region="RegionOne",
            profile="adminv3",
        )
        self.assertTrue(ret["keystone_|-testv3_|-testv3_|-endpoint_present"]["result"])

        # Region Two
        ret = self.run_state(
            "keystone.endpoint_present",
            name="testv3",
            description="Glance Service",
            interface="public",
            url="http://localhost:9292",
            region="RegionTwo",
            profile="adminv3",
        )
        self.assertTrue(ret["keystone_|-testv3_|-testv3_|-endpoint_present"]["result"])

        ret = self.run_state(
            "keystone.endpoint_present",
            name="testv3",
            description="Glance Service",
            interface="internal",
            url="http://localhost:9292",
            region="RegionTwo",
            profile="adminv3",
        )
        self.assertTrue(ret["keystone_|-testv3_|-testv3_|-endpoint_present"]["result"])

        ret = self.run_state(
            "keystone.endpoint_present",
            name="testv3",
            description="Glance Service",
            interface="admin",
            url="http://localhost:9292",
            region="RegionTwo",
            profile="adminv3",
        )
        self.assertTrue(ret["keystone_|-testv3_|-testv3_|-endpoint_present"]["result"])

        # Region One, change
        ret = self.run_state(
            "keystone.endpoint_present",
            name="testv3",
            description="Glance Service",
            interface="public",
            url="http://127.0.0.1:9292",
            region="RegionOne",
            profile="adminv3",
        )
        self.assertTrue(ret["keystone_|-testv3_|-testv3_|-endpoint_present"]["result"])

        ret = self.run_state(
            "keystone.endpoint_get",
            name="testv3",
            region="RegionOne",
            interface="public",
            profile="adminv3",
        )
        self.assertTrue(ret["keystone_|-testv3_|-testv3_|-endpoint_present"]["result"])
        self.assertTrue(
            ret["keystone_|-testv3_|-testv3_|-endpoint_present"]["endpoint"]["url"]
            == "http://127.0.0.1:9292"
        )

        ret = self.run_state(
            "keystone.project_present",
            name="testv3",
            description="Test v3 Tenant",
            profile="adminv3",
        )
        self.assertTrue(ret["keystone_|-testv3_|-testv3_|-project_present"]["result"])

        ret = self.run_state("keystone.role_present", name="user", profile="adminv3")
        self.assertTrue(ret["keystone_|-user_|-user_|-role_present"]["result"])

        ret = self.run_state(
            "keystone.user_present",
            name="testv3",
            email="testv3@example.com",
            project="testv3",
            password="testv3pass",
            roles={"testv3": ["user"]},
            profile="adminv3",
        )
        self.assertTrue(ret["keystone_|-testv3_|-testv3_|-user_present"]["result"])

        ret = self.run_state(
            "keystone.service_absent", name="testv3", profile="adminv3"
        )
        self.assertTrue(ret["keystone_|-testv3_|-testv3_|-service_absent"]["result"])



"""
    :codeauthor: Rahul Handay <rahulha@saltstack.com>
"""

import pytest

import salt.states.xmpp as xmpp
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {xmpp: {}}


def test_send_msg():
    """
    Test to send a message to an XMPP user
    """
    ret = {"name": "salt", "changes": {}, "result": None, "comment": ""}
    with patch.dict(xmpp.__opts__, {"test": True}):
        ret.update({"comment": "Need to send message to myaccount: salt"})
        assert xmpp.send_msg("salt", "myaccount", "salt@saltstack.com") == ret

    with patch.dict(xmpp.__opts__, {"test": False}):
        mock = MagicMock(return_value=True)
        with patch.dict(
            xmpp.__salt__, {"xmpp.send_msg": mock, "xmpp.send_msg_multi": mock}
        ):
            ret.update({"result": True, "comment": "Sent message to myaccount: salt"})
            assert xmpp.send_msg("salt", "myaccount", "salt@saltstack.com") == ret
